import utils
import os
import sys
import unittest
import subprocess

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.reader
import modelcif.util.make_mmcif  # Script should also be importable


MAKE_MMCIF = os.path.join(TOPDIR, 'modelcif', 'util', 'make_mmcif.py')


class Tests(unittest.TestCase):
    def test_simple(self):
        """Simple test of make_mmcif utility script"""
        incif = utils.get_input_file_name(TOPDIR, 'struct_only.cif')
        subprocess.check_call([sys.executable, MAKE_MMCIF, incif])
        with open('output.cif') as fh:
            s, = modelcif.reader.read(fh)
        self.assertEqual(s.title,
                         'Architecture of Pol II(G) and molecular mechanism '
                         'of transcription regulation by Gdown1')
        os.unlink('output.cif')

    def test_non_default_output(self):
        """Simple test of make_mmcif with non-default output name"""
        incif = utils.get_input_file_name(TOPDIR, 'struct_only.cif')
        subprocess.check_call([sys.executable, MAKE_MMCIF, incif,
                               'non-default-output.cif'])
        with open('non-default-output.cif') as fh:
            s, = modelcif.reader.read(fh)
        self.assertEqual(s.title,
                         'Architecture of Pol II(G) and molecular mechanism '
                         'of transcription regulation by Gdown1')
        os.unlink('non-default-output.cif')

    def test_no_title(self):
        """Check that make_mmcif adds missing title"""
        incif = utils.get_input_file_name(TOPDIR, 'no_title.cif')
        subprocess.check_call([sys.executable, MAKE_MMCIF, incif])
        with open('output.cif') as fh:
            s, = modelcif.reader.read(fh)
        self.assertEqual(s.title, 'Auto-generated system')
        os.unlink('output.cif')

    def test_bad_usage(self):
        """Bad usage of make_mmcif utility script"""
        ret = subprocess.call([sys.executable, MAKE_MMCIF])
        self.assertEqual(ret, 2)

    def test_same_file(self):
        """Check that make_mmcif fails if input and output are the same"""
        incif = utils.get_input_file_name(TOPDIR, 'struct_only.cif')
        ret = subprocess.call([sys.executable, MAKE_MMCIF, incif, incif])
        self.assertEqual(ret, 1)

    def test_not_modeled(self):
        """Check addition of not-modeled residue information"""
        incif = utils.get_input_file_name(TOPDIR, 'not_modeled.cif')
        subprocess.check_call([sys.executable, MAKE_MMCIF, incif])
        with open('output.cif') as fh:
            contents = fh.readlines()
        loop = contents.index("_pdbx_poly_seq_scheme.pdb_ins_code\n")
        scheme = "".join(contents[loop - 11:loop + 11])
        # Residues 5 and 6 in chain A, and 2 in chain B, are missing from
        # atom_site, so should now be missing from the scheme table.
        self.assertEqual(scheme, """#
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.mon_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.auth_seq_num
_pdbx_poly_seq_scheme.pdb_mon_id
_pdbx_poly_seq_scheme.auth_mon_id
_pdbx_poly_seq_scheme.pdb_strand_id
_pdbx_poly_seq_scheme.pdb_ins_code
A 1 1 VAL 2 2 VAL VAL A ?
A 1 2 GLY 3 3 GLY GLY A ?
A 1 3 GLN 4 4 GLN GLN A ?
A 1 4 GLN 5 5 GLN GLN A ?
A 1 5 TYR 5 ? ? ? A .
A 1 6 SER 6 ? ? ? A .
A 1 7 SER 8 8 SER SER A ?
B 2 1 ASP 3 3 ASP ASP B ?
B 2 2 GLU 2 ? ? ? B .
#
""")
        os.unlink('output.cif')


if __name__ == '__main__':
    unittest.main()



"""Classes to handle alignments between template structure(s)
   and target sequence(s).

   To create an alignment, first declare a class for the given kind of
   alignment by deriving from subclasses of :class:`AlignmentMode`
   (e.g. :class:`Global`) and :class:`AlignmentType` (e.g. :class:`Pairwise`).
   For example, a typical pairwise global alignment could be declared using::

       class Alignment(modelcif.alignment.Global, modelcif.alignment.Pairwise):
           pass
"""

import modelcif.data


class Identity:
    """Percent sequence identity between the template sequence and the target
       sequence being modeled.
       Use the correct subclass that corresponds to the denominator used
       when calculating the identity, for example
       :class:`ShorterSequenceIdentity`, or if the denominator is not covered
       here, subclass this class and provide a docstring to describe the
       denominator, e.g.::

           class CustomIdentity(Identity):
               "my custom sequence identity denominator"

       :param float value: The percent sequence identity value.
    """
    denominator = "Other"

    def __init__(self, value):
        self.value = value

    def _get_other_details(self):
        if (type(self) is not Identity
                and self.denominator == Identity.denominator):
            return self.__doc__.split('\n')[0]

    other_details = property(
        _get_other_details,
        doc="More information about a custom sequence identity denominator. "
            "By default it is the first line of the docstring.")


class ShorterSequenceIdentity(Identity):
    """Sequence identity calculated using the length of the shorter sequence
       as the denominator.
       See :class:`Identity` for more information."""
    other_details = None
    denominator = "Length of the shorter sequence"


class AlignedPositionsIdentity(Identity):
    """Sequence identity calculated using the number of aligned positions
       (including gaps) as the denominator.
       See :class:`Identity` for more information."""
    other_details = None
    denominator = "Number of aligned positions (including gaps)"


class AlignedResiduePairsIdentity(Identity):
    """Sequence identity calculated using the number of aligned residue pairs
       (not including gaps) as the denominator.
       See :class:`Identity` for more information."""
    other_details = None
    denominator = "Number of aligned residue pairs (not including the gaps)"


class MeanSequenceIdentity(Identity):
    """Sequence identity calculated using the arithmetic mean of the sequence
       lengths as the denominator.
       See :class:`Identity` for more information."""
    other_details = None
    denominator = "Arithmetic mean sequence length"


class Pair:
    """A single pairwise alignment between a single target and template chain.
       See :class:`AlignmentMode`. An alignment consists of one or more of
       these pairs.

       :param template: The template segment that is aligned, i.e. the
              seq_id range for the template and the sequence (including gaps)
              of one-letter codes.
       :type template: :class:`modelcif.TemplateSegment`
       :param target: The target segment that is aligned.
       :type target: output from :meth:`ihm.AsymUnit.segment`
       :param identity: The sequence identity between target and template,
              if known.
       :type identity: :class:`Identity`
       :param score: A measure of the quality of the alignment, if known.
       :type score: :class:`Score`
    """
    def __init__(self, template, target, identity=None, score=None):
        self.template, self.target, self.score = template, target, score
        self.identity = identity


class AlignmentMode(modelcif.data.Data):
    """Base class for all alignments. Actual alignments should derive
       from both a subclass of this class (e.g. :class:`Global`) and a
       subclass of :class:`AlignmentType`.

       :param str name: A short description of this alignment.
       :param pairs: List of individual target-template alignments.
       :type pairs: list of :class:`Pair` objects
       :param software: The software that was used to build the alignment.
       :type software: :class:`modelcif.Software`
             or :class:`modelcif.SoftwareGroup`
    """
    data_content_type = 'target-template alignment'

    def __init__(self, name, pairs, software=None):
        super().__init__(name)
        self.pairs = pairs
        self.software = software


class Global(AlignmentMode):
    """Base class for global alignments. See :class:`AlignmentMode` for
       more details."""
    mode = "global"


class Local(AlignmentMode):
    """Base class for local alignments. See :class:`AlignmentMode` for
       more details."""
    mode = "local"


class AlignmentType:
    """Base class for all alignment types. Actual alignments should derive
       from both a subclass of this class (e.g. :class:`Pairwise`) and a
       subclass of :class:`AlignmentMode`.
    """
    type = "other"


class Pairwise(AlignmentType):
    """An alignment between a single target and template.
       See :class:`AlignmentType` for more details."""
    type = "target-template pairwise alignment"
    other_details = None


class Multiple(AlignmentType):
    """A multiple sequence alignment between target and template.
       See :class:`AlignmentType` for more details."""
    type = "target-template MSA"
    other_details = None


class Score:
    """Base class for a quality score for a given target-template alignment.
       Usually a derived class such as :class:`BLASTEValue` is used, and
       passed to :class:`Pair` objects.

       :param float value: The actual score value.
    """
    type = "Other"

    def __init__(self, value):
        self.value = value


class BLASTEValue(Score):
    """BLAST e-value for an alignment. See :class:`Score` for more details."""
    type = "BLAST e-value"
    other_details = None


class HHblitsEValue(Score):
    """E-value computed by HHblits for an alignment. See :class:`Score` for
    more details."""
    type = "HHblits e-value"
    other_details = None



import utils
import os
import unittest
import io

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.reader
import modelcif.dumper


class Tests(unittest.TestCase):
    def test_model_with_assembly(self):
        """Test read of Model with Assembly followed by write"""
        sin = io.StringIO("""
loop_
_entity.id
_entity.type
_entity.pdbx_description
_entity.pdbx_number_of_molecules
_entity.formula_weight
_entity.details
1 polymer Nup84 2 100.0 .
#
loop_
_entity_poly_seq.entity_id
_entity_poly_seq.num
_entity_poly_seq.mon_id
_entity_poly_seq.hetero
1 1 ALA .
1 2 CYS .
#
loop_
_struct_asym.id
_struct_asym.entity_id
_struct_asym.details
A 1 'Model subunit A'
#
loop_
_ma_struct_assembly.ordinal_id
_ma_struct_assembly.assembly_id
_ma_struct_assembly.entity_id
_ma_struct_assembly.asym_id
_ma_struct_assembly.seq_id_begin
_ma_struct_assembly.seq_id_end
1 1 1 A 1 2
#
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' 1 4 'Homology model' .
#
loop_
_atom_site.group_PDB
_atom_site.id
_atom_site.type_symbol
_atom_site.label_atom_id
_atom_site.label_alt_id
_atom_site.label_comp_id
_atom_site.label_seq_id
_atom_site.auth_seq_id
_atom_site.pdbx_PDB_ins_code
_atom_site.label_asym_id
_atom_site.Cartn_x
_atom_site.Cartn_y
_atom_site.Cartn_z
_atom_site.occupancy
_atom_site.label_entity_id
_atom_site.auth_asym_id
_atom_site.B_iso_or_equiv
_atom_site.pdbx_PDB_model_num
ATOM 1 C CA . ALA 1 1 ? A 1.000 2.000 3.000 . 1 A . 1
""")
        s, = modelcif.reader.read(sin)
        sout = io.StringIO()
        modelcif.dumper.write(sout, [s])

    def test_model_without_assembly(self):
        """Test read of Model without Assembly followed by write"""
        sin = io.StringIO("""
loop_
_entity.id
_entity.type
_entity.pdbx_description
_entity.pdbx_number_of_molecules
_entity.formula_weight
_entity.details
1 polymer Nup84 2 100.0 .
#
loop_
_entity_poly_seq.entity_id
_entity_poly_seq.num
_entity_poly_seq.mon_id
_entity_poly_seq.hetero
1 1 ALA .
1 2 CYS .
#
loop_
_struct_asym.id
_struct_asym.entity_id
_struct_asym.details
A 1 'Model subunit A'
#
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' . 4 'Homology model' .
#
loop_
_atom_site.group_PDB
_atom_site.id
_atom_site.type_symbol
_atom_site.label_atom_id
_atom_site.label_alt_id
_atom_site.label_comp_id
_atom_site.label_seq_id
_atom_site.auth_seq_id
_atom_site.pdbx_PDB_ins_code
_atom_site.label_asym_id
_atom_site.Cartn_x
_atom_site.Cartn_y
_atom_site.Cartn_z
_atom_site.occupancy
_atom_site.label_entity_id
_atom_site.auth_asym_id
_atom_site.B_iso_or_equiv
_atom_site.pdbx_PDB_model_num
ATOM 1 C CA . ALA 1 1 ? A 1.000 2.000 3.000 . 1 A . 1
""")
        s, = modelcif.reader.read(sin)
        sout = io.StringIO()
        modelcif.dumper.write(sout, [s])


if __name__ == '__main__':
    unittest.main()



"""Classes for linking back to a sequence or structure database."""

import warnings
import ihm.reference
from ihm.reference import Alignment, SeqDif  # noqa: F401
from ihm.reference import InsertionSeqDif, DeletionSeqDif  # noqa: F401


class TargetReference(ihm.reference.Sequence):
    """Point to the sequence of a target :class:`modelcif.Entity` in a sequence
       database. Typically a subclass such as :class:`UniProt` is used,
       although to use a custom database, make a new subclass and provide
       a docstring to describe the database, e.g.::

           class CustomRef(TargetReference):
               "my custom database"

       Compare with :class:`modelcif.ReferenceDatabase`, which describes
       multiple sequences used in template searches or alignment
       construction; this class relates to just the modeled sequence itself.

       See also :attr:`alignments` to describe the correspondence between
       the database and entity sequences.

       :param str code: The name of the sequence in the database.
       :param str accession: The database accession.
       :param int align_begin: Beginning index of the sequence in the database.
              Deprecated; use :attr:`alignments` instead.
       :param int align_end: Ending index of the sequence in the database.
              Deprecated; use :attr:`alignments` instead.
       :param str isoform: Sequence isoform, if applicable.
       :param str ncbi_taxonomy_id: Taxonomy identifier provided by NCBI.
       :param str organism_scientific: Scientific name of the organism.
       :param sequence_version_date: Versioning date, e.g. for UniProtKB
                                     sequences this is usually the date of last
                                     modification from the DT line of an entry.
       :type sequence_version_date: :class:`datetime.date` or
                                    :class:`datetime.datetime`
       :param str sequence_crc64: The CRC64 sum of the original database
                                  sequence.
       :param str sequence: The complete database sequence, as a string of
              one-letter codes. If omitted, will default to the canonical
              sequence of the associated :class:`~modelcif.Entity`.
       :param str details: Longer text describing the sequence.
       :param bool is_primary: True iff this is the main input used in the
              modeling.
    """

    name = 'Other'

    def __init__(self, code, accession, align_begin=None, align_end=None,
                 isoform=None, ncbi_taxonomy_id=None,
                 organism_scientific=None, sequence_version_date=None,
                 sequence_crc64=None, sequence=None, details=None,
                 is_primary=None):
        super().__init__(
            db_name=self.name, db_code=code, accession=accession,
            sequence=sequence, details=details)
        self.align_begin, self.align_end = align_begin, align_end
        self.isoform = isoform
        self.ncbi_taxonomy_id = ncbi_taxonomy_id
        self.organism_scientific = organism_scientific
        self.sequence_version_date = sequence_version_date
        self.sequence_crc64 = sequence_crc64
        if align_begin or align_end:
            warnings.warn(
                "align_begin and align_end are deprecated, and will be "
                "removed in a future python-modelcif release. Specify the "
                "database sequence and provide one or more "
                "modelcif.reference.Alignment objects instead.", stacklevel=2)
        if sequence is None:
            warnings.warn(
                "No sequence provided. The canonical sequence of the Entity "
                "will be used instead.", stacklevel=2)
        #: All alignments between the reference and entity sequences, as
        #: :class:`Alignment` objects. If none are provided, a simple 1:1
        #: alignment is assumed.
        self.alignments = []
        self.is_primary = is_primary

    code = property(lambda self: self.db_code)

    def _get_other_details(self):
        if (type(self) is not TargetReference
                and self.name == TargetReference.name):
            return self.__doc__.split('\n')[0]

    other_details = property(
        _get_other_details,
        doc="More information about a custom reference type. "
            "By default it is the first line of the docstring.")


class UniProt(TargetReference):
    """Point to the sequence of an :class:`modelcif.Entity` in UniProt.

       These objects are typically passed to the :class:`modelcif.Entity`
       constructor for target sequences (for templates, see
       :class:`TemplateReference`).

       See :class:`TargetReference` for a description of the parameters.
    """
    name = 'UNP'
    other_details = None


class TemplateReference:
    """Point to the structure of a :class:`modelcif.Template` in a structure
       database.

       These objects are typically passed to the :class:`modelcif.Template`
       constructor for template sequences (for target sequences, see
       :class:`TargetReference`).

       Typically a subclass such as :class:`PDB` is used,
       although to use a custom database, make a new subclass and provide
       a docstring to describe the database, e.g.::

           class CustomRef(TemplateReference):
               "my custom database"

       :param str accession: The database accession.
       :param db_version_date: Versioning date, e.g. for PDB entries this is
                               usually the value of
                               ``_pdbx_audit_revision_history.revision_date``.
       :type db_version_date: :class:`datetime.date` or
                              :class:`datetime.datetime`
    """
    name = 'Other'

    def __init__(self, accession, db_version_date=None):
        self.accession = accession
        self.db_version_date = db_version_date

    def _get_other_details(self):
        if (type(self) is not TemplateReference
                and self.name == TemplateReference.name):
            return self.__doc__.split('\n')[0]

    other_details = property(
        _get_other_details,
        doc="More information about a custom reference type. "
            "By default it is the first line of the docstring.")


class PDB(TemplateReference):
    """Point to the structure of a :class:`modelcif.Template` in PDB.

       These objects are typically passed to the :class:`modelcif.Template`
       constructor.

       See :class:`TemplateReference` for a description of the parameters.
    """
    name = 'PDB'
    other_details = None


class AlphaFoldDB(TemplateReference):
    """Point to the structure of a :class:`modelcif.Template` in AlphaFold DB.

       These objects are typically passed to the :class:`modelcif.Template`
       constructor.

       See :class:`TemplateReference` for a description of the parameters.
    """
    name = 'AlphaFoldDB'
    other_details = None


class PubChem(TemplateReference):
    """Point to the structure of a :class:`modelcif.Template` in PubChem.

       These objects are typically passed to the :class:`modelcif.Template`
       constructor.

       See :class:`TemplateReference` for a description of the parameters. Use
       the PubChem CID as the accession code.
    """
    name = 'PubChem'
    other_details = None



# This example demonstrates writing an mmCIF file of a typical
# single-template homology or comparative model, including a ligand.
#
# This is very similar to the mkmodbase.py example; see that example for
# more details.

# Import used classes
import modelcif
import modelcif.model
import modelcif.dumper
import modelcif.reference
import modelcif.protocol
import modelcif.alignment
from modelcif.alignment import ShorterSequenceIdentity as SequenceIdentity
import ihm

system = modelcif.System(title='Ligand example')

# Describe the amino acid (polymer) sequences as Entity objects, for both
# template and model:
template_e = modelcif.Entity('AFVVTDNCIKCK', description='Template subunit')
model_e = modelcif.Entity('AYVINDSCIA', description='Model subunit')

# For non-polymers (e.g. ligands) we need to describe the chemistry of the
# ligand as a chemical component object, then create an Entity using that
# component. We only need to do this once because the ligand is the same
# in both template and model:
sf4 = ihm.NonPolymerChemComp("SF4", name='IRON/SULFUR CLUSTER',
                             formula='Fe4 S4')
ligand_e = modelcif.Entity([sf4], description='IRON/SULFUR CLUSTER')

# Create a Template for each chain (amino acids in chain A, ligand in chain B)
# and point to the original PDB, 5fd1:
s = modelcif.reference.PDB('5fd1')
templateA = modelcif.Template(
    entity=template_e, asym_id='A', model_num=1, name="Template polymer",
    transformation=modelcif.Transformation.identity(), references=[s])
templateB = modelcif.Template(
    entity=ligand_e, asym_id='B', model_num=1, name='Template ligand',
    transformation=modelcif.Transformation.identity(), references=[s])

# Define the model assembly, as two AsymUnits. NonPolymerFromTemplate is a
# subclass of AsymUnit that additionally notes the Template from which it
# was derived. In this case we state that the ligand was simply copied from
# the template into the target (explicit=False):
asymA = modelcif.AsymUnit(model_e, details='Model subunit A', id='A')
asymB = modelcif.NonPolymerFromTemplate(template=templateB, explicit=False,
                                        details='Model subunit B', id='B')
modeled_assembly = modelcif.Assembly((asymA, asymB), name='Modeled assembly')


# For the amino acid chain, add the modeling alignment, just as in the
# mkmodbase.py example:
class Alignment(modelcif.alignment.Global, modelcif.alignment.Pairwise):
    pass


p = modelcif.alignment.Pair(
    template=templateA.segment("AFVVTDNCIKCK", 1, 12),
    target=asymA.segment("AYVINDSC--IA", 1, 10),
    score=modelcif.alignment.BLASTEValue(1e-15),
    identity=SequenceIdentity(45.0))
aln = Alignment(name="Modeling alignment",
                pairs=[p])
system.alignments.append(aln)

# Add model coordinates, similarly to the mkmodbase.py example.
# Note that nonpolymers are not "sequences" and so seq_id=None.
atoms = [('A', 1, 'C', 'CA', 1., 2., 3.),
         ('A', 2, 'C', 'CA', 4., 5., 6.),
         ('A', 3, 'C', 'CA', 7., 8., 9.),
         ('B', None, 'FE', 'FE', 10., 10., 10.)]


class MyModel(modelcif.model.HomologyModel):
    asym_unit_map = {'A': asymA, 'B': asymB}

    def get_atoms(self):
        for asym, seq_id, type_symbol, atom_id, x, y, z in atoms:
            yield modelcif.model.Atom(
                asym_unit=self.asym_unit_map[asym], type_symbol=type_symbol,
                seq_id=seq_id, atom_id=atom_id, x=x, y=y, z=z,
                het=seq_id is None)


# Add the model and modeling protocol to the file and write them out:
model = MyModel(assembly=modeled_assembly, name='Best scoring model')

model_group = modelcif.model.ModelGroup([model], name='All models')
system.model_groups.append(model_group)

protocol = modelcif.protocol.Protocol()
protocol.steps.append(modelcif.protocol.ModelingStep(
    input_data=aln, output_data=model))
system.protocols.append(protocol)

with open('output.cif', 'w') as fh:
    modelcif.dumper.write(fh, [system])



"""Classes to annotate models with quality scores.

   To use, first declare a class for the desired score by deriving
   from both a subclass of :class:`MetricMode` (which defines the part
   of the system the metric applies to) and a subclass of :class:`MetricType`
   (which describes the meaning of the score value). Set the ``software``
   attribute to point to the software used to calculate the metric
   (as a :class:`modelcif.SoftwareGroup` or :class:`modelcif.Software` object).
   For example to declare a global distance score::

       class MyScore(modelcif.qa_metric.Global, modelcif.qa_metric.Distance):
           "My distance-based quality score"
           software = modelcif.Software(...)

   The name and description of the score in the mmCIF file will be taken from
   the name and docstring of the Python class, unless the
   :attr:`MetricMode.name` or :attr:`MetricMode.description` attributes are
   overridden in the subclass.

   QA metric objects should be added to
   :attr:`modelcif.model.Model.qa_metrics`.
"""

from ihm.util import _text_choice_property


class MetricMode:
    """Base class for the mode of a quality metric.
       Use a derived class such as :class:`Global`, :class:`Local`,
       :class:`LocalPairwise`, :class:`Feature`, :class:`FeaturePairwise`,
       or :class:`Dihedral` for declaring a new score.
    """
    name = property(lambda x: type(x).__name__,
                    doc="Short name of this score. By default it is just the "
                        "class name, but this can be overridden in subclasses "
                        "(for example to create names containing spaces).")

    description = property(lambda x: x.__doc__.split("\n")[0],
                           doc="Longer text description of this score. By "
                               "default it is the first line of the "
                               "docstring.")


class Global(MetricMode):
    """A score that is calculated per-model.

       :param float value: The score value (see :class:`MetricType`).
    """

    mode = "global"

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "<%s(value=%r)>" % (type(self).__name__, self.value)


class Local(MetricMode):
    """A score that is calculated on a single residue.

       :param residue: The residue that is scored.
       :type residue: :class:`modelcif.Residue`
       :param float value: The score value (see :class:`MetricType`).
    """

    mode = "local"

    def __init__(self, residue, value):
        self.residue = residue
        self.value = value

    def __repr__(self):
        return "<%s(residue=%r, value=%r)>" % (type(self).__name__,
                                               self.residue, self.value)


class LocalPairwise(MetricMode):
    """A score that is calculated between two residues.

       :param residue1: The first residue that is scored.
       :type residue1: :class:`modelcif.Residue`
       :param residue2: The second residue that is scored.
       :type residue2: :class:`modelcif.Residue`
       :param float value: The score value (see :class:`MetricType`).
    """

    mode = "local-pairwise"

    def __init__(self, residue1, residue2, value):
        self.residue1 = residue1
        self.residue2 = residue2
        self.value = value

    def __repr__(self):
        return ("<%s(residue1=%r, residue2=%r, value=%r)>"
                % (type(self).__name__, self.residue1, self.residue2,
                   self.value))


class Feature(MetricMode):
    """A score that is calculated on a single feature.

       :param feature: The feature that is scored.
       :type feature: :class:`modelcif.Feature`
       :param float value: The score value (see :class:`MetricType`).
    """

    mode = "per-feature"

    def __init__(self, feature, value):
        self.feature = feature
        self.value = value

    def __repr__(self):
        return "<%s(feature=%r, value=%r)>" % (type(self).__name__,
                                               self.feature, self.value)
    _all_features = property(lambda self: (self.feature,))


class FeaturePairwise(MetricMode):
    """A score that is calculated between two features.

       :param feature1: The first feature that is scored.
       :type feature1: :class:`modelcif.Feature`
       :param feature2: The second feature that is scored.
       :type feature2: :class:`modelcif.Feature`
       :param float value: The score value (see :class:`MetricType`).
    """

    mode = "per-feature-pair"

    def __init__(self, feature1, feature2, value):
        self.feature1 = feature1
        self.feature2 = feature2
        self.value = value

    def __repr__(self):
        return ("<%s(feature1=%r, feature2=%r, value=%r)>"
                % (type(self).__name__, self.feature1, self.feature2,
                   self.value))
    _all_features = property(lambda self: (self.feature1, self.feature2))


class Dihedral(MetricMode):
    """A score that is calculated on a dihedral.

       :param int atom_id_1: The first atom ID in the dihedral.
       :param int atom_id_2: The second atom ID in the dihedral.
       :param int atom_id_3: The third atom ID in the dihedral.
       :param int atom_id_4: The fourth atom ID in the dihedral.
       :param float value: The score value (see :class:`MetricType`).
       :param str quality: Outcome or result of the analysis.
       :param str smarts_pattern: Optional SMARTS pattern that specifies the
              dihedral angle in its chemical environment; or that defines the
              fragment used for dihedral scanning.
    """

    mode = "dihedral"

    def __init__(self, atom_id_1, atom_id_2, atom_id_3, atom_id_4, value,
                 quality, smarts_pattern=None):
        self.atom_id_1, self.atom_id_2 = atom_id_1, atom_id_2
        self.atom_id_3, self.atom_id_4 = atom_id_3, atom_id_4
        self.value = value
        self.quality, self.smarts_pattern = quality, smarts_pattern

    quality = _text_choice_property(
        "quality", ["relaxed", "tolerable", "strained"],
        doc="Outcome or result of the analysis.")

    def __repr__(self):
        return ("<%s(atom_id_1=%r, atom_id_2=%r, atom_id_3=%r, atom_id_4=%r, "
                "value=%r)>" % (type(self).__name__, self.atom_id_1,
                                self.atom_id_2, self.atom_id_3, self.atom_id_4,
                                self.value))


class MetricType:
    """Base class for the type of a quality metric.
       Generally a derived class such as :class:`ZScore` or :class:`Distance`
       is used to declare a new score, but a custom type can also be declared
       by deriving from this class and providing a docstring to describe
       the metric type::

           class MPQSMetricType(modelcif.qa_metric.MetricType):
                "composite score, values >1.1 are reliable"
    """

    type = "other"

    def _get_other_details(self):
        # Find most derived class of MetricType before we pulled in MetricMode
        # and use the first line of its docstring as other_details
        if self.type == MetricType.type:
            for base in type(self).mro():
                if (issubclass(base, MetricType)
                        and base is not MetricType
                        and not issubclass(base, MetricMode)):
                    return base.__doc__.split('\n')[0]

    other_details = property(
        _get_other_details,
        doc="More information about this metric type. By default it is the "
            "first line of the MetricType subclass docstring.")


class ZScore(MetricType):
    """Score that is the number of standard deviations from optimal/best.
       See :class:`MetricType` for more information."""
    type = "zscore"
    other_details = None


class Energy(MetricType):
    """Energy score (the lower the energy, the better the quality).
       See :class:`MetricType` for more information."""
    type = "energy"
    other_details = None


class Distance(MetricType):
    """Distance score (the lower the distance, the better the quality).
       See :class:`MetricType` for more information."""
    type = "distance"
    other_details = None


class NormalizedScore(MetricType):
    """Normalized score ranging from 0 to 1.
       See :class:`MetricType` for more information."""
    type = "normalized score"
    other_details = None


class PAE(MetricType):
    """Score that is a predicted aligned error.
       See :class:`MetricType` for more information."""
    type = "PAE"
    other_details = None


class ContactProbability(MetricType):
    """Score that is a contact probability of a pairwise interaction.
       See :class:`MetricType` for more information."""
    type = "contact probability"
    other_details = None


class PLDDT(MetricType):
    """Predicted lDDT-CA score in [0,100] (higher score, means better
       accuracy). See :class:`MetricType` for more information."""
    type = "pLDDT"
    other_details = None


class PLDDT01(MetricType):
    """Predicted lDDT-CA score in [0,1] (higher score, means better
       accuracy). See :class:`MetricType` for more information."""
    type = "pLDDT in [0,1]"
    other_details = None


class PLDDTAllAtom(MetricType):
    """Predicted lDDT all atom score in [0,100] (higher score, means better
       accuracy). See :class:`MetricType` for more information."""
    type = "pLDDT all-atom"
    other_details = None


class PLDDTAllAtom01(MetricType):
    """Predicted lDDT all atom score in [0,1] (higher score, means better
       accuracy). See :class:`MetricType` for more information."""
    type = "pLDDT all-atom in [0,1]"
    other_details = None


class PLDDTToPolymer(MetricType):
    """Predicted lDDT with distances from each atom to CA or C1' of nearby
       polymer residues [0,100] (higher score, means better accuracy).
       See :class:`MetricType` for more information."""
    type = "pLDDT to polymer"
    other_details = None


class PTM(MetricType):
    """Predicted TM-score in [0,1] (higher value means higher confidence).
    See :class:`MetricType` for more information."""
    type = "pTM"
    other_details = None


class IpTM(MetricType):
    """Protein-protein interface score, based on TM-score in [0,1].
    See :class:`MetricType` for more information."""
    type = "ipTM"
    other_details = None


class Boolean(MetricType):
    """0 or 1 depending on whether a check passed.
    See :class:`MetricType` for more information."""
    type = "boolean"
    other_details = None



import modelcif
import modelcif.dumper
import modelcif.reader
import os
import unittest


class Tests(unittest.TestCase):
    def test_basic(self):
        """Basic install test"""
        system = modelcif.System(title='test system')

        entity_a = modelcif.Entity('AAA', description='Subunit A')
        entity_b = modelcif.Entity('AAAAAA', description='Subunit B')
        system.entities.extend((entity_a, entity_b))

        # Test output in mmCIF format
        with open('output.cif', 'w') as fh:
            modelcif.dumper.write(fh, [system])

        # Make sure we can read back the file
        with open('output.cif') as fh:
            sys2, = modelcif.reader.read(fh)
        self.assertEqual(sys2.title, 'test system')

        os.unlink('output.cif')


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import utils

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif
import modelcif.protocol
import modelcif.descriptor
import modelcif.associated
import ihm


class Tests(unittest.TestCase):
    def test_all_data(self):
        """Test _all_data() method"""
        s = modelcif.System()
        e1 = modelcif.Entity("D")
        e2 = modelcif.Entity("M")
        s.entities.extend((e1, e2))

        e3 = modelcif.Entity("A")
        s.data.extend((e1, e3))

        s.data_groups.append('something not a group')
        e4 = modelcif.Entity("M")
        s.data_groups.append(modelcif.data.DataGroup([e1, e4]))

        e5 = modelcif.Entity("M")
        f = modelcif.associated.File(path='foo', details='bar', data=e5)
        s.repositories.append(modelcif.associated.Repository('/', [f]))

        d = s._all_data()
        # List may contain duplicates
        self.assertEqual(list(d), [e1, e3, e1, e2, e1, e4, e5])

    def test_all_asym_units(self):
        """Test _all_asym_units() method"""
        s = modelcif.System()
        e1 = modelcif.Entity("DDDD")
        e2 = modelcif.Entity("MMMM")
        a1 = modelcif.AsymUnit(e1)
        a2 = modelcif.AsymUnit(e2)
        s.asym_units.append(a1)

        asmb = modelcif.Assembly((a1, a2(1, 2)))
        s.assemblies.append(asmb)

        asyms = s._all_asym_units()
        # List may contain duplicates and should be all AsymUnit,
        # not AsymUnitRange
        self.assertEqual(list(asyms), [a1, a1, a2])

    def test_all_entities(self):
        """Test _all_entities() method"""
        s = modelcif.System()
        e1 = modelcif.Entity("DDDD")
        e2 = modelcif.Entity("MMMM")
        s.entities.append(e1)

        a1 = modelcif.AsymUnit(e1)
        s.asym_units.append(a1)

        t2 = modelcif.Template(e2, asym_id='A', model_num=1,
                               transformation=None)
        s.templates.append(t2)

        es = s._all_entities()
        # List may contain duplicates, but does not contain template entity e2
        self.assertEqual(list(es), [e1, e1])

    def test_all_data_groups(self):
        """Test _all_data_groups() method"""
        s = modelcif.System()
        e1 = modelcif.Entity("A")
        s.data_groups.append(e1)
        e2 = modelcif.Entity("C")

        p = modelcif.protocol.Protocol()
        p.steps.append(modelcif.protocol.ModelingStep(
            input_data=e1, output_data=e2))
        p.steps.append(modelcif.protocol.ModelingStep(
            input_data=None, output_data=None))
        s.protocols.append(p)

        d = s._all_data_groups()
        self.assertEqual(list(d), [e1, e1, e2])

    def test_all_template_transformations(self):
        """Test _all_template_transformations() method"""
        s = modelcif.System()
        tr1 = 'tr1'
        tr2 = 'tr2'
        s.template_transformations.extend((tr1, tr2))

        template = modelcif.Template('mockentity', asym_id="A", model_num=1,
                                     name="test template",
                                     transformation=tr1)
        s.templates.append(template)

        tt = s._all_template_transformations()
        # List may contain duplicates
        self.assertEqual(list(tt), [tr1, tr2, tr1])

    def test_transformation(self):
        """Test Transformation class"""
        _ = modelcif.Transformation([[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                                    [1, 2, 3])

    def test_identity_transformation(self):
        """Test identity transformation"""
        t = modelcif.Transformation.identity()
        for i in range(3):
            self.assertAlmostEqual(t.tr_vector[i], 0., delta=0.1)
            for j in range(3):
                self.assertAlmostEqual(t.rot_matrix[i][j],
                                       1. if i == j else 0., delta=0.1)

        # Should always get the same object
        t2 = modelcif.Transformation.identity()
        self.assertIs(t, t2)

    def test_all_software_groups(self):
        """Test _all_software_groups() method"""
        s = modelcif.System()
        sg1 = 'sg1'
        sg2 = 'sg2'
        s.software_groups.extend((sg1, sg2))

        p = modelcif.protocol.Protocol()
        p.steps.append(modelcif.protocol.ModelingStep(
            input_data=None, output_data=None, software=sg1))
        s.protocols.append(p)

        allsg = s._all_software_groups()
        # List may contain duplicates
        self.assertEqual(list(allsg), [sg1, sg2, sg1])

    def test_all_ref_software(self):
        """Test _all_ref_software() method"""
        s1 = modelcif.Software(
            name='foo', version='1.0',
            classification='1', description='2', location='3')
        s2 = modelcif.Software(
            name='foo', version='2.0',
            classification='4', description='5', location='6')
        p = modelcif.SoftwareParameter(name='foo', value='bar')
        s2param = modelcif.SoftwareWithParameters(s2, [p])
        s = modelcif.System()
        s.software_groups.append(modelcif.SoftwareGroup((s1, s2param)))
        s.software_groups.append(s1)

        e1 = modelcif.Entity("DDDD")
        t1 = modelcif.Template(e1, asym_id='A', model_num=1,
                               transformation=None)
        s.templates.append(t1)

        # Old-style ChemComp without descriptors
        c1 = ihm.NonPolymerChemComp('C1', name='C1')
        if hasattr(c1, 'descriptors'):
            del c1.descriptors

        # ChemComp with descriptor without software
        c2 = ihm.NonPolymerChemComp('C2', name='C2')
        c2.descriptors = [modelcif.descriptor.IUPACName('foo')]

        # ChemComp with descriptor with software
        c3 = ihm.NonPolymerChemComp('C3', name='C3')
        s3 = modelcif.Software(
            name='foo', version='2.0',
            classification='4', description='5', location='6')
        c3.descriptors = [modelcif.descriptor.IUPACName('foo', software=s3)]

        e2 = modelcif.Entity([c1, c2, c3])
        s.entities.append(e2)

        alls = s._all_ref_software()
        # List may contain duplicates
        self.assertEqual(list(alls), [s1, s2, s1, s3])

    def test_software_parameter(self):
        """Test SoftwareParameter class"""
        p = modelcif.SoftwareParameter(name='foo', value=42)
        self.assertEqual(p.name, 'foo')
        self.assertEqual(p.value, 42)
        self.assertIsNone(p.description)
        _ = repr(p)

    def test_software_with_parameters(self):
        """Test SoftwareWithParameters class"""
        s = modelcif.Software(
            name='foo', version='1.0',
            classification='1', description='2', location='3')
        p = modelcif.SoftwareParameter(name='foo', value=42)
        swp = modelcif.SoftwareWithParameters(software=s, parameters=[p])
        self.assertEqual(swp.software.name, 'foo')
        self.assertEqual(swp.parameters, [p])
        self.assertEqual(swp.name, 'foo')
        self.assertEqual(swp.classification, '1')
        self.assertEqual(swp.description, '2')
        self.assertEqual(swp.location, '3')
        self.assertEqual(swp.type, 'program')
        self.assertEqual(swp.version, '1.0')
        self.assertIsNone(swp.citation)

    def test_template(self):
        """Test Template class"""
        e1 = modelcif.Entity("DDDD")
        t1 = modelcif.Template(e1, asym_id='A', model_num=1,
                               transformation=None)
        self.assertEqual(t1.seq_id_range, (1, 4))
        self.assertEqual(t1.template, t1)

    def test_software_group_parameters(self):
        """Test old-style SoftwareGroup construction with parameters"""
        s = modelcif.Software(
            name='foo', version='1.0',
            classification='1', description='2', location='3')
        p = modelcif.SoftwareParameter(name='foo', value=42)
        self.assertWarns(UserWarning,
                         modelcif.SoftwareGroup, [s], parameters=[p])

    def test_feature(self):
        """Test Feature base class"""
        f = modelcif.Feature()
        self.assertIs(f._get_entity_type(), ihm.unknown)


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.descriptor


class Tests(unittest.TestCase):
    def test_descriptor(self):
        """Test Descriptor classes"""
        base = modelcif.descriptor.Descriptor("1abc")
        self.assertEqual(base.value, "1abc")
        _ = repr(base)


if __name__ == '__main__':
    unittest.main()



# This example demonstrates using the library to make an mmCIF file of a
# typical single-template single-chain homology or comparative model, similar
# to those generated by MODELLER (https://salilab.org/modeller/) and deposited
# in the ModBase database (https://modbase.compbio.ucsf.edu/)

# For a more complete (but less documented) script to convert a complete
# ModBase PDB file into a corresponding mmCIF or BinaryCIF file, see
# https://github.com/salilab/modbase_utils/blob/main/modbase_pdb_to_cif.py

# Import used classes.
import modelcif
import modelcif.protocol
import modelcif.model
import modelcif.dumper
import modelcif.reference
import modelcif.qa_metric
import modelcif.alignment
# Different methods measure "sequence identity" in different ways, so import
# the class that matches the way Modeller understands it (number of identical
# aligned residues, divided by the length of the shorter sequence)
from modelcif.alignment import ShorterSequenceIdentity as SequenceIdentity
import ihm.citations
import modelcif.reader

# First, we create a system, which contains everything we know about the
# modeling. A single mmCIF file can contain multiple Systems, but in most
# cases we use just one:
system = modelcif.System(title='S54091 hypothetical protein YPR070w')

# List the authors of this file (here these are the ModBase authors)
system.authors.extend(('Pieper U', 'Webb B', 'Narayanan E', 'Sali A'))

# Describe the software that was used in the modeling
modpipe_software = modelcif.Software(
    name='ModPipe', classification='comparative modeling',
    location='https://salilab.org/modpipe/', type='program',
    version='SVN.r1703', description='Comparative modeling pipeline')
# Every object we create must ultimately be linked to the System, which
# maintains simple lists for each type of object. For example, there is a
# list system.software (like system.authors above) which can be used for
# any Software object not referenced by any other object. But in this case
# we're going to use these Software objects further on in the script, so
# don't need to explicitly add them here.

modeller_software = modelcif.Software(
    name='MODELLER', classification='comparative modeling',
    location='https://salilab.org/modeller/', type='program',
    version='SVN', citation=ihm.citations.modeller,
    description='Comparative modeling by satisfaction of spatial restraints')

# Next, we define "entities", unique sequences in the system, as Entity
# objects. First, the template sequence:
template_e = modelcif.Entity('DMACDTFIKCC', description='Template subunit')

# Next, the target (model) sequence, together with a link to the reference
# sequence (in UniProt):
s = modelcif.reference.UniProt(code='MED1_YEAST', accession='Q12321',
                               sequence='DSYVETLDCC')
model_e = modelcif.Entity('DSYVETLDCC', description='Model subunit',
                          references=[s])

# Next, we define asymmetric units for everything we modeled.
# These roughly correspond to chains in a traditional PDB file. Multiple
# asymmetric units may map to the same entity (for example if there are
# several copies of a given protein).
asymA = modelcif.AsymUnit(model_e, details='Model subunit A', id='A')

# Next, we group asymmetric units into assemblies.
modeled_assembly = modelcif.Assembly((asymA,), name='Modeled assembly')

# In a similar fashion, we declare a Template for each chain that we used
# as a template structure, with a link to the reference structure database
# (PDB).
s = modelcif.reference.PDB('3nc1')
template = modelcif.Template(
    entity=template_e, asym_id='A', model_num=1, name="Template Structure",
    transformation=modelcif.Transformation.identity(),
    references=[s])


# Now, we describe the alignment between target and template.
# python-ma provides various subclasses to use here. All ModBase structures
# use a simple pairwise global alignment between target and template, so
# declare a suitable class:
class Alignment(modelcif.alignment.Global, modelcif.alignment.Pairwise):
    pass


# An alignment consists of a list of aligned target-template segments.
# Here we provide the residue ranges and the actual alignment, including gaps,
# between the two, together with the sequence identity and any score available
# for the alignment (here we have the BLAST e-value):
p = modelcif.alignment.Pair(
    template=template.segment("DMACDTFIK", 1, 9),
    target=asymA.segment("DSYV-ETLD", 1, 8),
    score=modelcif.alignment.BLASTEValue(1e-15),
    identity=SequenceIdentity(45.0))
aln = Alignment(name="Modeling alignment", software=modpipe_software,
                pairs=[p])
# Alignments aren't used by any objects; they should be added directly
# to the System:
system.alignments.append(aln)

# For the actual model coordinates, we must subclass a suitable class and
# override the get_atoms() method to return a list of Atom objects. This design
# avoids having a separate copy of every atom in memory.
# Modeller models are comparative or homology models, so we subclass
# HomologyModel. For the purposes of this example, we just return a simple
# static list of atoms:

atoms = [('A', 1, 'C', 'CA', 1., 2., 3.),
         ('A', 2, 'C', 'CA', 4., 5., 6.),
         ('A', 3, 'C', 'CA', 7., 8., 9.),
         ('A', 4, 'C', 'CA', 10., 11., 12.)]


class MyModel(modelcif.model.HomologyModel):
    # Map our asym unit names to ModelCIF asym_unit objects:
    asym_unit_map = {'A': asymA}

    def get_atoms(self):
        for asym, seq_id, type_symbol, atom_id, x, y, z in atoms:
            yield modelcif.model.Atom(
                asym_unit=self.asym_unit_map[asym], type_symbol=type_symbol,
                seq_id=seq_id, atom_id=atom_id, x=x, y=y, z=z)


# Link the model to the Assembly that describes all subunits
model = MyModel(assembly=modeled_assembly, name='Best scoring model')

# Next, we describe the modeling protocol:
protocol = modelcif.protocol.Protocol()
protocol.steps.append(modelcif.protocol.TemplateSearchStep(
    name='ModPipe Seq-Prf (0001)', software=modpipe_software,
    input_data=model_e, output_data=aln))
protocol.steps.append(modelcif.protocol.ModelingStep(
    software=modeller_software, input_data=aln, output_data=model))
protocol.steps.append(modelcif.protocol.ModelSelectionStep(
    software=modpipe_software, input_data=model, output_data=model))
# Protocols aren't used by any other objects; they should be added directly
# to the System:
system.protocols.append(protocol)


# We can also attach quality scores to our model(s). To do this we must
# first define the scores by creating subclasses using a MetricMode
# (e.g. global, per-residue) and a MetricType (e.g. distance, z-score).
# Here we define the quality scores used by the ModPipe pipeline that is used
# by ModBase. Note that one score (MPQS) uses a custom metric type, while
# another (zDOPE) is a simple global z-score:
class MPQSMetricType(modelcif.qa_metric.MetricType):
    """composite score, values >1.1 are considered reliable"""


class MPQS(modelcif.qa_metric.Global, MPQSMetricType):
    """ModPipe Quality Score"""
    software = modpipe_software


class zDOPE(modelcif.qa_metric.Global, modelcif.qa_metric.ZScore):
    """Normalized DOPE"""
    software = modeller_software


class TSVModRMSD(modelcif.qa_metric.Global, modelcif.qa_metric.Distance):
    """TSVMod predicted RMSD (MSALL)"""
    software = None


class TSVModNO35(modelcif.qa_metric.Global,
                 modelcif.qa_metric.NormalizedScore):
    """TSVMod predicted native overlap (MSALL)"""
    software = None


# Add qa metrics to the model
model.qa_metrics.extend((MPQS(0.853452), zDOPE(0.31), TSVModRMSD(12.996),
                         TSVModNO35(0.143)))

# All ModBase QA metrics are global, but the library also supports per-residue
# or pairwise (between two residues) scores. Here's a fictional example for a
# z-score on the 4th residue of the first chain in the model, and a distance
# score between the 1st and 3rd residues:


class SomeLocalScore(modelcif.qa_metric.Local, modelcif.qa_metric.ZScore):
    """A per-residue z-score"""
    software = None


class SomePairScore(modelcif.qa_metric.LocalPairwise,
                    modelcif.qa_metric.Distance):
    """A distance score between two residues"""
    software = None


model.qa_metrics.append(SomeLocalScore(asymA.residue(4), -0.1))
model.qa_metrics.append(SomePairScore(asymA.residue(1), asymA.residue(3), 1.0))

# Models should be grouped together using ModelGroup and then added to the
# top-level System. Here we only have a single model in the group:
model_group = modelcif.model.ModelGroup([model], name='All models')
system.model_groups.append(model_group)

# Once the system is complete, we can write it out to an mmCIF file:
with open('output.cif', 'w') as fh:
    modelcif.dumper.write(fh, [system])

# We can also *read* an mmCIF file and create a set of Python objects from it.
# Here we read in the file we just created:
with open('output.cif') as fh:
    s, = modelcif.reader.read(fh)
for t in s.templates:
    print(t.name, "-".join(c.id for c in t.entity.sequence))
for e in s.entities:
    print(e.description, "-".join(c.id for c in e.sequence))



from datetime import date
import warnings
import utils
import os
import unittest
from io import StringIO
try:
    import msgpack
except ImportError:
    msgpack = None

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.dumper
import modelcif.protocol
import modelcif.model
import modelcif.reference
import modelcif.alignment
import modelcif.associated
import modelcif.descriptor
import ihm.format
import ihm.dumper


def _get_dumper_output(dumper, system, check=True):
    dumper._check = check
    fh = StringIO()
    writer = ihm.format.CifWriter(fh)
    dumper.dump(system, writer)
    return fh.getvalue()


class Tests(unittest.TestCase):
    def test_write(self):
        """Test write() function"""
        sys1 = modelcif.System(id='system1')
        sys2 = modelcif.System(id='system 2+3')
        fh = StringIO()
        modelcif.dumper.write(fh, [sys1, sys2])
        lines = fh.getvalue().split('\n')
        self.assertEqual(lines[:2], ["data_system1", "_entry.id system1"])
        if lines[9] == 'data_system23':
            self.assertEqual(lines[9:11],
                             ["data_system23", "_entry.id 'system 2+3'"])
        else:
            self.assertEqual(lines[11:13],
                             ["data_system23", "_entry.id 'system 2+3'"])

    def test_audit_conform_dumper(self):
        """Test AuditConformDumper"""
        system = modelcif.System()
        dumper = modelcif.dumper._AuditConformDumper()
        out = _get_dumper_output(dumper, system)
        lines = sorted(out.split('\n'))
        self.assertEqual(lines[1].split()[0], "_audit_conform.dict_location")
        self.assertEqual(lines[2].rstrip('\r\n'),
                         "_audit_conform.dict_name mmcif_ma.dic")
        self.assertEqual(lines[3].split()[0], "_audit_conform.dict_version")

    def test_database_dumper(self):
        """Test DatabaseDumper"""
        system = modelcif.System()
        dumper = modelcif.dumper._DatabaseDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, '')

        system = modelcif.System(
            database=modelcif.Database(id='foo', code='bar'))
        dumper = modelcif.dumper._DatabaseDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, "_database_2.database_code bar\n"
                              "_database_2.database_id foo\n")

    def test_software_group_dumper(self):
        """Test SoftwareGroupDumper"""
        class MockObject:
            pass
        p1 = modelcif.SoftwareParameter(name='foo', value=42)
        p2 = modelcif.SoftwareParameter(name='bar', value=True)
        p3 = modelcif.SoftwareParameter(name='baz', value='ok')
        intlist = modelcif.SoftwareParameter(name='intlist', value=[1, 2, 3])
        floatlist = modelcif.SoftwareParameter(
            name='floatlist', value=(1., 2., 3.))
        mixlist = modelcif.SoftwareParameter(name='mixlist', value=[1, 2., 3])
        s1 = modelcif.Software(
            name='s1', classification='test code',
            description='Some test program',
            version=1, location='http://test.org')
        s1._id = 1
        s2 = modelcif.Software(
            name='s2', classification='test code',
            description='Some test program',
            version=1, location='http://test.org')
        s2._id = 2
        s3 = modelcif.Software(
            name='s3', classification='test code',
            description='Some test program',
            version=1, location='http://test.org')
        s3._id = 3
        system = modelcif.System()
        aln1 = MockObject()
        aln1.pairs = []
        aln1.software = modelcif.SoftwareGroup((s1, s2))
        # SoftwareGroup.parameters should be ignored
        aln1.software.parameters.append('garbage')
        aln2 = MockObject()
        aln2.pairs = []
        aln2.software = s3
        aln3 = MockObject()
        aln3.pairs = []
        s3param = modelcif.SoftwareWithParameters(
            software=s3, parameters=[p1, p2, p3, intlist, floatlist, mixlist])
        aln3.software = modelcif.SoftwareGroup((s2, s3param))

        # Duplicate parameters, should get the same ID as for aln3
        aln4 = MockObject()
        aln4.pairs = []
        aln4.software = modelcif.SoftwareGroup((s3param,))

        system.alignments.extend((aln1, aln2, aln3, aln4))
        system._before_write()  # populate system.software_groups
        dumper = modelcif.dumper._SoftwareGroupDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        # Should have two groups (s1, s2) and (s2, s3) and another
        # singleton group (s3)
        self.assertEqual(out, """#
loop_
_ma_software_parameter.parameter_id
_ma_software_parameter.group_id
_ma_software_parameter.data_type
_ma_software_parameter.name
_ma_software_parameter.value
_ma_software_parameter.description
1 1 integer foo 42 .
2 1 boolean bar YES .
3 1 string baz ok .
4 1 integer-csv intlist 1,2,3 .
5 1 float-csv floatlist 1.0,2.0,3.0 .
6 1 float-csv mixlist 1,2.0,3 .
#
#
loop_
_ma_software_group.ordinal_id
_ma_software_group.group_id
_ma_software_group.software_id
_ma_software_group.parameter_group_id
1 1 1 .
2 1 2 .
3 2 3 .
4 3 2 .
5 3 3 1
6 4 3 1
#
""")

    def test_bad_software_parameter(self):
        """Test invalid SoftwareParameter"""
        p1 = modelcif.SoftwareParameter(name='foo', value=['string', 'list'])
        s1 = modelcif.Software(
            name='s1', classification='test code',
            description='Some test program',
            version=1, location='http://test.org')
        system = modelcif.System()
        system.software.append(s1)
        s1param = modelcif.SoftwareWithParameters(s1, parameters=[p1])
        sg1 = modelcif.SoftwareGroup([s1param])
        system.software_groups.append(sg1)
        dumper = modelcif.dumper._SoftwareGroupDumper()
        dumper.finalize(system)
        # Only lists of ints or floats are supported, not strings
        self.assertRaises(TypeError, _get_dumper_output, dumper, system)

    def test_data_dumper(self):
        """Test DataDumper"""
        system = modelcif.System()
        entity = modelcif.Entity("DMA", description='test entity')
        system.entities.append(entity)
        # Template and target use same entity here (but different data IDs)
        template = modelcif.Template(
            entity, asym_id="A", model_num=1, name="test template",
            transformation=modelcif.Transformation.identity())
        system.templates.append(template)
        system.data.append(modelcif.data.Data(name="test other",
                                              details="test details"))
        system._before_write()  # populate system.data
        dumper = modelcif.dumper._DataDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_data.id
_ma_data.name
_ma_data.content_type
_ma_data.content_type_other_details
1 'test other' other 'test details'
2 'test template' 'template structure' .
3 'test entity' target .
#
""")

    def test_data_group_dumper(self):
        """Test DataGroupDumper"""
        system = modelcif.System()
        tgt_e1 = modelcif.Entity("D")
        tgt_e2 = modelcif.Entity("M")
        tgt_e3 = modelcif.Entity("A")
        tgt_e1._data_id = 1
        tgt_e2._data_id = 2
        tgt_e3._data_id = 3
        system.entities.extend((tgt_e1, tgt_e2, tgt_e3))
        dg12 = modelcif.data.DataGroup((tgt_e1, tgt_e2))
        p = modelcif.protocol.Protocol()
        p.steps.append(modelcif.protocol.ModelingStep(
            input_data=dg12, output_data=tgt_e3))
        system.protocols.append(p)
        system._before_write()  # populate system.data_groups
        dumper = modelcif.dumper._DataGroupDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        # First group (tgt_e1,tgt_e2); second group contains just tgt_e3
        self.assertEqual(out, """#
loop_
_ma_data_group.ordinal_id
_ma_data_group.group_id
_ma_data_group.data_id
1 1 1
2 1 2
3 2 3
#
""")

    def test_data_ref_db_dumper(self):
        """Test DataRefDBDumper"""
        system = modelcif.System()
        system.data.append(modelcif.ReferenceDatabase(
            name='testdb', url='testurl', version='1.0',
            release_date=date(1979, 11, 22)))
        system.data.append(modelcif.data.Data(name="test other",
                                              details="test details"))
        dumper = modelcif.dumper._DataDumper()
        dumper.finalize(system)  # Assign Data IDs
        dumper = modelcif.dumper._DataRefDBDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_data_ref_db.data_id
_ma_data_ref_db.name
_ma_data_ref_db.location_url
_ma_data_ref_db.version
_ma_data_ref_db.release_date
1 testdb testurl 1.0 1979-11-22
#
""")

    def test_qa_metric_dumper(self):
        """Test QAMetricDumper"""
        system = modelcif.System()
        s1 = modelcif.Software(
            name='s1', classification='test code',
            description='Some test program',
            version=1, location='http://test.org')
        s1._group_id = 1

        class MockObject:
            pass

        class CustomMetricType(modelcif.qa_metric.MetricType):
            """my custom type"""

        class DistanceScore(modelcif.qa_metric.Global,
                            modelcif.qa_metric.Distance):
            """test description"""
            name = "test score"
            software = s1

        class CustomScore(modelcif.qa_metric.Global, CustomMetricType):
            """Description does not match docstring"""
            description = "custom description"
            software = None

        class LocalScore(modelcif.qa_metric.Local, modelcif.qa_metric.ZScore):
            """custom local description
               Second line of docstring (ignored)"""
            name = "custom local score"
            software = None

        class PairScore(modelcif.qa_metric.LocalPairwise,
                        modelcif.qa_metric.Energy):
            """custom pair description"""
            name = "custom pair score"
            software = None

        class FeatureScore(modelcif.qa_metric.Feature,
                           modelcif.qa_metric.ZScore):
            """feature score"""
            name = "feature score"
            software = None

        class FeaturePairwiseScore(modelcif.qa_metric.FeaturePairwise,
                                   modelcif.qa_metric.ZScore):
            """feature pairwise score"""
            name = "feature pairwise score"
            software = None

        class DihedralScore(modelcif.qa_metric.Dihedral,
                            modelcif.qa_metric.Energy):
            """dihedral score"""
            name = "dihedral score"
            software = None

        m1 = DistanceScore(42.)
        m2 = CustomScore(99.)
        m3 = DistanceScore(60.)
        e1 = modelcif.Entity('ACGT')
        asym = modelcif.AsymUnit(e1, 'foo')
        asym._id = 'Z'
        m4 = LocalScore(asym.residue(2), 20.)
        m5 = PairScore(asym.residue(1), asym.residue(3), 30.)
        resf = modelcif.PolyResidueFeature((asym.residue(1), asym.residue(2)))
        instf = modelcif.EntityInstanceFeature((asym,))
        m6 = FeatureScore(resf, 40.)
        m7 = FeaturePairwiseScore(resf, instf, 50.)
        m8 = DihedralScore(1, 2, 3, 4, 60., "tolerable", "some-smarts")
        model = MockObject()
        model._id = 18
        model.qa_metrics = [m1, m2, m3, m4, m5, m6, m7, m8]
        mg = modelcif.model.ModelGroup((model,))
        system.model_groups.append(mg)
        # Assign feature IDs
        dumper = modelcif.dumper._FeatureDumper()
        dumper.finalize(system)
        dumper = modelcif.dumper._QAMetricDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_qa_metric.id
_ma_qa_metric.name
_ma_qa_metric.description
_ma_qa_metric.type
_ma_qa_metric.mode
_ma_qa_metric.type_other_details
_ma_qa_metric.software_group_id
1 'test score' 'test description' distance global . 1
2 CustomScore 'custom description' other global 'my custom type' .
3 'custom local score' 'custom local description' zscore local . .
4 'custom pair score' 'custom pair description' energy local-pairwise . .
5 'feature score' 'feature score' zscore per-feature . .
6 'feature pairwise score' 'feature pairwise score' zscore per-feature-pair . .
7 'dihedral score' 'dihedral score' energy dihedral . .
#
#
loop_
_ma_qa_metric_global.ordinal_id
_ma_qa_metric_global.model_id
_ma_qa_metric_global.metric_id
_ma_qa_metric_global.metric_value
1 18 1 42.000
2 18 2 99.000
3 18 1 60.000
#
#
loop_
_ma_qa_metric_local.ordinal_id
_ma_qa_metric_local.model_id
_ma_qa_metric_local.label_asym_id
_ma_qa_metric_local.label_seq_id
_ma_qa_metric_local.label_comp_id
_ma_qa_metric_local.metric_id
_ma_qa_metric_local.metric_value
1 18 Z 2 CYS 3 20.000
#
#
loop_
_ma_qa_metric_local_pairwise.ordinal_id
_ma_qa_metric_local_pairwise.model_id
_ma_qa_metric_local_pairwise.label_asym_id_1
_ma_qa_metric_local_pairwise.label_seq_id_1
_ma_qa_metric_local_pairwise.label_comp_id_1
_ma_qa_metric_local_pairwise.label_asym_id_2
_ma_qa_metric_local_pairwise.label_seq_id_2
_ma_qa_metric_local_pairwise.label_comp_id_2
_ma_qa_metric_local_pairwise.metric_id
_ma_qa_metric_local_pairwise.metric_value
1 18 Z 1 ALA Z 3 GLY 4 30.000
#
#
loop_
_ma_qa_metric_feature.ordinal_id
_ma_qa_metric_feature.model_id
_ma_qa_metric_feature.feature_id
_ma_qa_metric_feature.metric_id
_ma_qa_metric_feature.metric_value
1 18 1 5 40.000
#
#
loop_
_ma_qa_metric_feature_pairwise.ordinal_id
_ma_qa_metric_feature_pairwise.model_id
_ma_qa_metric_feature_pairwise.feature_id_1
_ma_qa_metric_feature_pairwise.feature_id_2
_ma_qa_metric_feature_pairwise.metric_id
_ma_qa_metric_feature_pairwise.metric_value
1 18 1 2 6 50.000
#
#
loop_
_ma_qa_metric_dihedral.ordinal_id
_ma_qa_metric_dihedral.atom_id_1
_ma_qa_metric_dihedral.atom_id_2
_ma_qa_metric_dihedral.atom_id_3
_ma_qa_metric_dihedral.atom_id_4
_ma_qa_metric_dihedral.metric_id
_ma_qa_metric_dihedral.metric_value
_ma_qa_metric_dihedral.quality
_ma_qa_metric_dihedral.smarts_pattern
1 1 2 3 4 7 60.000 tolerable some-smarts
#
""")

    def test_feature_dumper(self):
        """Test FeatureDumper"""
        system = modelcif.System()

        class MockObject:
            pass

        class TestScore(modelcif.qa_metric.Feature, modelcif.qa_metric.ZScore):
            """test score"""
            name = "test score"
            software = None

        e1 = modelcif.Entity('ACGT')
        asym = modelcif.AsymUnit(e1, 'foo')
        e2 = ihm.Entity([ihm.NonPolymerChemComp('HEM')])
        heme = modelcif.AsymUnit(e2, 'heme')

        asym._id = 'Y'
        heme._id = 'Z'
        atomf = modelcif.AtomFeature((1, 2, 3), details='atom f')
        resf = modelcif.PolyResidueFeature((asym.residue(1), asym.residue(2)),
                                           details='prf')
        instf = modelcif.EntityInstanceFeature((asym,))
        inst2f = modelcif.EntityInstanceFeature((heme,))
        atoms = TestScore(atomf, 20.)
        ress = TestScore(resf, 30.)
        insts = TestScore(instf, 40.)
        inst2s = TestScore(inst2f, 40.)

        model = MockObject()
        model._id = 18
        model.qa_metrics = [atoms, ress, insts, inst2s]
        mg = modelcif.model.ModelGroup((model,))
        system.model_groups.append(mg)

        dumper = modelcif.dumper._FeatureDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_feature_list.feature_id
_ma_feature_list.feature_type
_ma_feature_list.entity_type
_ma_feature_list.details
1 atom other 'atom f'
2 residue polymer prf
3 'entity instance' polymer .
4 'entity instance' non-polymer .
#
#
loop_
_ma_atom_feature.ordinal_id
_ma_atom_feature.feature_id
_ma_atom_feature.atom_id
1 1 1
2 1 2
3 1 3
#
#
loop_
_ma_poly_residue_feature.ordinal_id
_ma_poly_residue_feature.feature_id
_ma_poly_residue_feature.label_asym_id
_ma_poly_residue_feature.label_seq_id
_ma_poly_residue_feature.label_comp_id
1 2 Y 1 ALA
2 2 Y 2 CYS
#
#
loop_
_ma_entity_instance_feature.ordinal_id
_ma_entity_instance_feature.feature_id
_ma_entity_instance_feature.label_asym_id
1 3 Y
2 4 Z
#
""")
        # Test empty feature
        emptyf = modelcif.EntityInstanceFeature(())
        emptys = TestScore(emptyf, 20.)
        model.qa_metrics = [emptys]
        dumper = modelcif.dumper._FeatureDumper()
        dumper.finalize(system)
        self.assertRaises(ValueError, _get_dumper_output, dumper, system)
        # Should work with checks disabled
        _ = _get_dumper_output(dumper, system, check=False)

        # Test feature that selects multiple entity types
        multf = modelcif.EntityInstanceFeature((asym, heme))
        mults = TestScore(multf, 20.)
        model.qa_metrics = [mults]
        dumper = modelcif.dumper._FeatureDumper()
        dumper.finalize(system)
        self.assertRaises(ValueError, _get_dumper_output, dumper, system)
        # Should work with checks disabled
        _ = _get_dumper_output(dumper, system, check=False)

    def test_protocol_dumper(self):
        """Test ProtocolDumper"""
        class MockObject:
            pass
        indat = MockObject()
        indat._data_group_id = 1
        outdat = MockObject()
        outdat._data_group_id = 2
        system = modelcif.System()
        s1 = modelcif.Software(
            name='s1', classification='test code',
            description='Some test program',
            version=1, location='http://test.org')
        s1._group_id = 42
        p = modelcif.protocol.Protocol()
        p.steps.append(modelcif.protocol.TemplateSearchStep(
            name='tsstep', details="some details", software=s1,
            input_data=indat, output_data=outdat))
        p.steps.append(modelcif.protocol.ModelingStep(
            name='modstep', input_data=indat, output_data=outdat))
        p.steps.append(modelcif.protocol.ModelingStep(
            name='nullstep', input_data=None, output_data=None))
        system.protocols.append(p)
        dumper = modelcif.dumper._ProtocolDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_protocol_step.ordinal_id
_ma_protocol_step.protocol_id
_ma_protocol_step.step_id
_ma_protocol_step.method_type
_ma_protocol_step.step_name
_ma_protocol_step.details
_ma_protocol_step.software_group_id
_ma_protocol_step.input_data_group_id
_ma_protocol_step.output_data_group_id
1 1 1 'template search' tsstep 'some details' 42 1 2
2 1 2 modeling modstep . . 1 2
3 1 3 modeling nullstep . . . .
#
""")

    def test_model_dumper(self):
        """Test ModelDumper"""
        class CustomModel(modelcif.model.Model):
            """custom model"""

        system = modelcif.System()
        e1 = modelcif.Entity('ACGT')
        e1._id = 9
        system.entities.append(e1)
        asym = modelcif.AsymUnit(e1, 'foo')
        asym._id = 'A'
        system.asym_units.append(asym)
        asmb = modelcif.Assembly((asym,))
        asmb._id = 2
        model1 = modelcif.model.HomologyModel(assembly=asmb, name='test model')
        model1._data_id = 42
        model1._atoms = [modelcif.model.Atom(asym_unit=asym, seq_id=1,
                                             atom_id='C', type_symbol='C',
                                             x=1.0, y=2.0, z=3.0)]
        model2 = modelcif.model.AbInitioModel(assembly=asmb, name='model2')
        model2._data_id = 43
        model3 = CustomModel(assembly=asmb, name='model3')
        model3._data_id = 44
        mg = modelcif.model.ModelGroup((model1, model2, model3),
                                       name='test group')
        system.model_groups.append(mg)
        # model1 is in both groups
        mg = modelcif.model.ModelGroup((model1,),
                                       name='second group')
        # ModelGroup constructor only supports details with python-ihm > 1.8
        mg.details = 'second group details'
        system.model_groups.append(mg)
        dumper = modelcif.dumper._ModelDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_name
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 'test model' 42 'Homology model' .
2 model2 43 'Ab initio model' .
3 model3 44 Other 'custom model'
#
#
loop_
_ma_model_group.id
_ma_model_group.name
_ma_model_group.details
1 'test group' .
2 'second group' 'second group details'
#
#
loop_
_ma_model_group_link.group_id
_ma_model_group_link.model_id
1 1
1 2
1 3
2 1
#
#
loop_
_atom_site.group_PDB
_atom_site.id
_atom_site.type_symbol
_atom_site.label_atom_id
_atom_site.label_alt_id
_atom_site.label_comp_id
_atom_site.label_seq_id
_atom_site.auth_seq_id
_atom_site.pdbx_PDB_ins_code
_atom_site.label_asym_id
_atom_site.Cartn_x
_atom_site.Cartn_y
_atom_site.Cartn_z
_atom_site.occupancy
_atom_site.label_entity_id
_atom_site.auth_asym_id
_atom_site.auth_comp_id
_atom_site.B_iso_or_equiv
_atom_site.pdbx_PDB_model_num
ATOM 1 C C . ALA 1 1 ? A 1.000 2.000 3.000 . 9 A ALA . 1
#
#
loop_
_atom_type.symbol
C
#
""")

    def test_poly_seq_scheme_dumper(self):
        """Test PolySeqSchemeDumper with ModelCIF models"""
        system = modelcif.System()
        e1 = modelcif.Entity('ACGT')
        e1._id = 9
        system.entities.append(e1)
        asym = modelcif.AsymUnit(e1, 'foo')
        asym._id = 'A'
        system.asym_units.append(asym)
        asmb = modelcif.Assembly((asym,))
        asmb._id = 2
        model1 = modelcif.model.HomologyModel(assembly=asmb, name='test model')
        model1._data_id = 42
        model1._atoms = [modelcif.model.Atom(asym_unit=asym, seq_id=1,
                                             atom_id='C', type_symbol='C',
                                             x=1.0, y=2.0, z=3.0)]
        mg = modelcif.model.ModelGroup((model1,),
                                       name='test group')
        # Add at least one model, since the PolySeqSchemeDumper checks all
        # models' not_modeled_residue_ranges member when writing the table
        system.model_groups.append(mg)
        dumper = ihm.dumper._PolySeqSchemeDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.mon_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.auth_seq_num
_pdbx_poly_seq_scheme.pdb_mon_id
_pdbx_poly_seq_scheme.auth_mon_id
_pdbx_poly_seq_scheme.pdb_strand_id
_pdbx_poly_seq_scheme.pdb_ins_code
A 9 1 ALA 1 1 ALA ALA A .
A 9 2 CYS 2 2 CYS CYS A .
A 9 3 GLY 3 3 GLY GLY A .
A 9 4 THR 4 4 THR THR A .
#
""")

    def test_target_ref_db_dumper(self):
        """Test TargetRefDBDumper"""

        class CustomRef(modelcif.reference.TargetReference):
            """my custom ref"""

        system = modelcif.System()
        # Default alignment but with explicit align begin, end
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ref1 = modelcif.reference.UniProt(
                code='testcode', accession='testacc', align_begin=4,
                align_end=8, isoform='testiso', ncbi_taxonomy_id='1234',
                organism_scientific='testorg',
                sequence_version_date=date(1979, 11, 22),
                sequence_crc64="A123B456C789D1E2",
                sequence='ACGT',
                is_primary=True)
        # Default alignment (entire sequence)
        ref2 = modelcif.reference.UniProt(code='c2', accession='a2',
                                          sequence='ACGT')
        ref3 = CustomRef(code='c3', accession='a3', isoform=ihm.unknown,
                         sequence='ACGT')

        # Explicit alignment that extends to the end of the db sequence
        ref4 = modelcif.reference.UniProt(code='c4', accession='a4',
                                          sequence='CCACGT')
        ref4.alignments.append(modelcif.reference.Alignment(db_begin=3))

        # Explicit alignment with explicit db_end
        ref5 = modelcif.reference.UniProt(code='c5', accession='a5',
                                          sequence='XXXACXXGTXXX',
                                          is_primary=False)
        ref5.alignments.append(modelcif.reference.Alignment(
            db_begin=4, db_end=5))
        ref5.alignments.append(modelcif.reference.Alignment(
            db_begin=8, db_end=9))

        e1 = modelcif.Entity('ACGT', references=[ref1, ref2, ref3, ref4, ref5])
        e1._id = 1
        system.entities.append(e1)

        dumper = modelcif.dumper._TargetRefDBDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_target_ref_db_details.target_entity_id
_ma_target_ref_db_details.db_name
_ma_target_ref_db_details.db_name_other_details
_ma_target_ref_db_details.db_code
_ma_target_ref_db_details.db_accession
_ma_target_ref_db_details.seq_db_isoform
_ma_target_ref_db_details.seq_db_align_begin
_ma_target_ref_db_details.seq_db_align_end
_ma_target_ref_db_details.ncbi_taxonomy_id
_ma_target_ref_db_details.organism_scientific
_ma_target_ref_db_details.seq_db_sequence_version_date
_ma_target_ref_db_details.seq_db_sequence_checksum
_ma_target_ref_db_details.is_primary
1 UNP . testcode testacc testiso 4 8 1234 testorg 1979-11-22 A123B456C789D1E2
YES
1 UNP . c2 a2 . 1 4 . . . . .
1 Other 'my custom ref' c3 a3 ? 1 4 . . . . .
1 UNP . c4 a4 . 3 6 . . . . .
1 UNP . c5 a5 . 4 9 . . . . NO
#
""")

    def test_alignment_dumper(self):
        """Test AlignmentDumper"""

        class CustomRef(modelcif.reference.TemplateReference):
            """my custom ref"""

        class Alignment(modelcif.alignment.Global,
                        modelcif.alignment.Pairwise):
            pass

        class LocalAlignment(modelcif.alignment.Local,
                             modelcif.alignment.Multiple):
            pass

        system = modelcif.System()
        tmp_e = modelcif.Entity('ACG')
        tgt_e = modelcif.Entity('ACE')
        tgt_e._id = 1
        system.entities.extend((tmp_e, tgt_e))
        asym = modelcif.AsymUnit(tgt_e, id='A')
        asym._id = 'A'
        system.asym_units.append(asym)
        ref1 = modelcif.reference.PDB('1abc',
                                      db_version_date=date(1979, 11, 22))
        ref2 = CustomRef('2xyz')
        ref3 = modelcif.reference.PubChem("1234")
        ref4 = modelcif.reference.AlphaFoldDB("P12345",
                                              db_version_date=date(2022, 6, 1))
        tr = modelcif.Transformation.identity()
        tr._id = 42
        t = modelcif.Template(tmp_e, asym_id='H', model_num=1, name='testtmp',
                              transformation=tr,
                              references=[ref1, ref2, ref3, ref4],
                              strand_id='Z')
        t._data_id = 99
        p = modelcif.alignment.Pair(
            template=t.segment('AC-G', 1, 3),
            target=asym.segment('ACE-', 1, 3),
            score=modelcif.alignment.BLASTEValue("1e-15"),
            identity=modelcif.alignment.ShorterSequenceIdentity(42.))
        aln = Alignment(name='testaln', pairs=[p])
        aln._data_id = 100
        system.alignments.append(aln)
        # The same alignment using HHblits e-value
        p1 = modelcif.alignment.Pair(
            template=p.template,
            target=p.target,
            score=modelcif.alignment.HHblitsEValue("1e-14"),
            identity=p.identity)
        # The same alignment with missing score and identity
        p2 = modelcif.alignment.Pair(
            template=p.template,
            target=p.target)
        aln = Alignment(name='testaln', pairs=[p1, p2])
        aln._data_id = 101
        system.alignments.append(aln)
        # Local alignment with no pairs
        aln2 = LocalAlignment(name='testaln2', pairs=[])
        aln2._data_id = 102
        system.alignments.append(aln2)
        system._before_write()  # populate system.templates

        dumper = modelcif.dumper._AlignmentDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' polymer 42 99 A H . 1 Z
2 1 'reference database' polymer 42 99 A H . 1 Z
3 1 'reference database' polymer 42 99 A H . 1 Z
#
#
loop_
_ma_template_poly.template_id
_ma_template_poly.seq_one_letter_code
_ma_template_poly.seq_one_letter_code_can
1 ACG ACG
#
#
loop_
_ma_template_poly_segment.id
_ma_template_poly_segment.template_id
_ma_template_poly_segment.residue_number_begin
_ma_template_poly_segment.residue_number_end
1 1 1 3
#
#
loop_
_ma_template_ref_db_details.template_id
_ma_template_ref_db_details.db_name
_ma_template_ref_db_details.db_name_other_details
_ma_template_ref_db_details.db_accession_code
_ma_template_ref_db_details.db_version_date
1 PDB . 1abc 1979-11-22
1 Other 'my custom ref' 2xyz .
1 PubChem . 1234 .
1 AlphaFoldDB . P12345 2022-06-01
#
#
loop_
_ma_target_template_poly_mapping.id
_ma_target_template_poly_mapping.template_segment_id
_ma_target_template_poly_mapping.target_asym_id
_ma_target_template_poly_mapping.target_seq_id_begin
_ma_target_template_poly_mapping.target_seq_id_end
1 1 A 1 3
2 1 A 1 3
3 1 A 1 3
#
#
loop_
_ma_alignment_info.alignment_id
_ma_alignment_info.data_id
_ma_alignment_info.software_group_id
_ma_alignment_info.alignment_length
_ma_alignment_info.alignment_type
_ma_alignment_info.alignment_mode
1 100 . 4 'target-template pairwise alignment' global
2 101 . 4 'target-template pairwise alignment' global
3 102 . . 'target-template MSA' local
#
#
loop_
_ma_alignment_details.ordinal_id
_ma_alignment_details.alignment_id
_ma_alignment_details.template_segment_id
_ma_alignment_details.target_asym_id
_ma_alignment_details.score_type
_ma_alignment_details.score_type_other_details
_ma_alignment_details.score_value
_ma_alignment_details.percent_sequence_identity
_ma_alignment_details.sequence_identity_denominator
_ma_alignment_details.sequence_identity_denominator_other_details
1 1 1 A 'BLAST e-value' . 1e-15 42.000 'Length of the shorter sequence' .
2 2 1 A 'HHblits e-value' . 1e-14 42.000 'Length of the shorter sequence' .
3 2 1 A . . . . . .
#
#
loop_
_ma_alignment.ordinal_id
_ma_alignment.alignment_id
_ma_alignment.target_template_flag
_ma_alignment.sequence
1 1 1 ACE-
2 1 2 AC-G
3 2 1 ACE-
4 2 2 AC-G
5 2 1 ACE-
6 2 2 AC-G
#
""")

    def test_non_poly_template_unused(self):
        """Test AlignmentDumper with unused nonpolymeric template"""
        system = modelcif.System()
        # Polymeric entity
        e1 = ihm.Entity('ACGT')
        t1 = modelcif.Template(
            e1, asym_id="A", model_num=1, name="test template",
            transformation=modelcif.Transformation.identity(),
            entity_id=9)
        t1._id = 1
        t1._data_id = 99
        # Non-polymeric entity
        e2 = ihm.Entity([ihm.NonPolymerChemComp('HEM')], description='heme')
        t2 = modelcif.Template(
            e2, asym_id="B", model_num=1, name="test template",
            transformation=modelcif.Transformation.identity(),
            entity_id=10)
        t2._id = 2
        t2._data_id = 100
        system.templates.extend((t1, t2))
        dumper = modelcif.dumper._AlignmentDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' polymer 42 99 . A 9 1 A
2 2 'reference database' non-polymer 42 100 . B 10 1 B
#
#
loop_
_ma_template_poly.template_id
_ma_template_poly.seq_one_letter_code
_ma_template_poly.seq_one_letter_code_can
1 ACGT ACGT
#
#
loop_
_ma_template_non_poly.template_id
_ma_template_non_poly.comp_id
_ma_template_non_poly.details
2 HEM heme
#
""")

    def test_non_poly_template_used(self):
        """Test AlignmentDumper with used nonpolymeric template"""
        system = modelcif.System()
        # Polymeric entity
        e1 = ihm.Entity('ACGT')
        t1 = modelcif.Template(
            e1, asym_id="A", model_num=1, name="test template",
            transformation=modelcif.Transformation.identity())
        t1._id = 1
        t1._data_id = 98
        # Non-polymeric entity
        e2 = ihm.Entity([ihm.NonPolymerChemComp('HEM')], description='heme')
        # Template should use entity_id, not e2._id
        e2._id = "THIS SHOULD BE IGNORED"
        t2 = modelcif.Template(
            e2, asym_id="B", model_num=1, name="test template",
            transformation=modelcif.Transformation.identity(),
            entity_id=9)
        t2._id = 2
        t2._data_id = 99
        system.templates.extend((t1, t2))

        a2 = modelcif.NonPolymerFromTemplate(template=t2, explicit=True)
        a2._id = 'X'
        system.asym_units.append(a2)

        dumper = modelcif.dumper._AlignmentDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 2 'reference database' non-polymer 42 99 X B 9 1 B
2 1 'reference database' polymer 42 98 . A . 1 A
#
#
loop_
_ma_template_poly.template_id
_ma_template_poly.seq_one_letter_code
_ma_template_poly.seq_one_letter_code_can
1 ACGT ACGT
#
#
loop_
_ma_template_non_poly.template_id
_ma_template_non_poly.comp_id
_ma_template_non_poly.details
2 HEM heme
#
""")

    def test_custom_template_unused(self):
        """Test AlignmentDumper with custom template"""
        system = modelcif.System()
        e1 = ihm.Entity('ACGT')
        t1 = modelcif.CustomTemplate(
            e1, asym_id="A", model_num=1, name="test template",
            transformation=modelcif.Transformation.identity(),
            entity_id=9, details='my custom template')
        t1.atoms.append(modelcif.TemplateAtom(
            seq_id=1, atom_id='CA',
            type_symbol='C', x=0.0, y=1.0, z=2.0, occupancy=0.5,
            biso=2.0, charge=1.0, auth_seq_id=42, auth_comp_id='XXX',
            auth_atom_id='X'))
        t1.atoms.append(modelcif.TemplateAtom(
            seq_id=2, atom_id='OXT',
            type_symbol='O', x=1.0, y=2.0, z=3.0))
        t1._id = 1
        t1._data_id = 99
        system.templates.append(t1)
        dumper = modelcif.dumper._AlignmentDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 customized polymer 42 99 . A 9 1 A
#
#
loop_
_ma_template_poly.template_id
_ma_template_poly.seq_one_letter_code
_ma_template_poly.seq_one_letter_code_can
1 ACGT ACGT
#
#
loop_
_ma_template_customized.template_id
_ma_template_customized.details
1 'my custom template'
#
#
loop_
_ma_template_coord.template_id
_ma_template_coord.group_PDB
_ma_template_coord.ordinal_id
_ma_template_coord.type_symbol
_ma_template_coord.label_atom_id
_ma_template_coord.label_comp_id
_ma_template_coord.label_seq_id
_ma_template_coord.label_asym_id
_ma_template_coord.auth_seq_id
_ma_template_coord.auth_asym_id
_ma_template_coord.auth_atom_id
_ma_template_coord.auth_comp_id
_ma_template_coord.Cartn_x
_ma_template_coord.Cartn_y
_ma_template_coord.Cartn_z
_ma_template_coord.occupancy
_ma_template_coord.label_entity_id
_ma_template_coord.B_iso_or_equiv
_ma_template_coord.formal_charge
1 ATOM 1 C CA ALA 1 A 42 A X XXX 0.000 1.000 2.000 0.500 9 2.000 1.000
1 ATOM 2 O OXT CYS 2 A . A . . 1.000 2.000 3.000 . 9 . .
#
""")

    def test_template_transform_dumper(self):
        """Test TemplateTransformDumper"""
        system = modelcif.System()
        tr1 = modelcif.Transformation(
            rot_matrix=[[-0.64, 0.09, 0.77], [0.76, -0.12, 0.64],
                        [0.15, 0.99, 0.01]],
            tr_vector=[1., 2., 3.])
        system.template_transformations.append(tr1)
        dumper = modelcif.dumper._TemplateTransformDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_template_trans_matrix.id
_ma_template_trans_matrix.rot_matrix[1][1]
_ma_template_trans_matrix.rot_matrix[2][1]
_ma_template_trans_matrix.rot_matrix[3][1]
_ma_template_trans_matrix.rot_matrix[1][2]
_ma_template_trans_matrix.rot_matrix[2][2]
_ma_template_trans_matrix.rot_matrix[3][2]
_ma_template_trans_matrix.rot_matrix[1][3]
_ma_template_trans_matrix.rot_matrix[2][3]
_ma_template_trans_matrix.rot_matrix[3][3]
_ma_template_trans_matrix.tr_vector[1]
_ma_template_trans_matrix.tr_vector[2]
_ma_template_trans_matrix.tr_vector[3]
1 -0.640000 0.760000 0.150000 0.090000 -0.120000 0.990000 0.770000 0.640000
0.010000 1.000 2.000 3.000
#
""")

    def test_target_entity_dumper(self):
        """Test TargetEntityDumper"""
        system = modelcif.System()
        e1 = modelcif.Entity("D")
        e1._id = 42
        e1._data_id = 99
        system.entities.append(e1)

        a1 = modelcif.AsymUnit(e1, 'foo')
        a1._id = 'X'
        system.asym_units.append(a1)

        dumper = modelcif.dumper._TargetEntityDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_target_entity.entity_id
_ma_target_entity.data_id
_ma_target_entity.origin
42 99 designed
#
#
loop_
_ma_target_entity_instance.asym_id
_ma_target_entity_instance.entity_id
_ma_target_entity_instance.details
X 42 foo
#
""")

    def test_associated_dumper(self):
        """Test AssociatedDumper"""
        system = modelcif.System()
        e = modelcif.Entity('M')
        # File in a repository
        f1 = modelcif.associated.File(path='foo.txt', details='test file')
        # File in an archive
        f2 = modelcif.associated.File(path='bar.txt', details='test file2')
        zf = modelcif.associated.ZipFile(path='t.zip', files=[f2])
        # Local file with data
        f3 = modelcif.associated.File(path='baz.txt', details='test file3',
                                      data=e)
        f4 = modelcif.associated.QAMetricsFile(path='baz.txt',
                                               details='test file4')
        r = modelcif.associated.Repository(url_root='https://example.com',
                                           files=[f1, zf])
        r2 = modelcif.associated.Repository(url_root=None, files=[f3, f4])
        system.repositories.extend((r, r2))

        system._before_write()  # populate data
        dumper = modelcif.dumper._DataDumper()
        dumper.finalize(system)  # Assign Data IDs

        dumper = modelcif.dumper._AssociatedDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_entry_associated_files.id
_ma_entry_associated_files.entry_id
_ma_entry_associated_files.file_url
_ma_entry_associated_files.file_type
_ma_entry_associated_files.file_format
_ma_entry_associated_files.file_content
_ma_entry_associated_files.details
_ma_entry_associated_files.data_id
1 model https://example.com/foo.txt file other other 'test file' .
2 model https://example.com/t.zip archive zip 'archive with multiple files' . .
3 model baz.txt file other other 'test file3' 1
4 model baz.txt file cif 'QA metrics' 'test file4' .
#
#
loop_
_ma_associated_archive_file_details.id
_ma_associated_archive_file_details.archive_file_id
_ma_associated_archive_file_details.file_path
_ma_associated_archive_file_details.file_format
_ma_associated_archive_file_details.file_content
_ma_associated_archive_file_details.description
_ma_associated_archive_file_details.data_id
1 2 bar.txt other other 'test file2' .
#
""")

        # Should be an error to put a zip file inside another zip
        zf2 = modelcif.associated.ZipFile(path='test2.zip', files=[])
        zf.files.append(zf2)
        self.assertRaises(ValueError, dumper.finalize, system)

    def test_write_associated(self):
        """Test write() function with associated files"""
        s = modelcif.System(id='system1')

        f = modelcif.associated.CIFFile(
            path='test_write_associated.cif',
            categories=['struct', '_AUDIT_CONFORM'],
            entry_details='test details', entry_id='testcif')
        f2 = modelcif.associated.File(path='foo.txt', details='test file')
        r = modelcif.associated.Repository(url_root='https://example.com',
                                           files=[f, f2])
        s.repositories.append(r)

        fh = StringIO()
        modelcif.dumper.write(fh, [s])
        main_file = fh.getvalue()
        with open('test_write_associated.cif') as fh:
            assoc_file = fh.read()
        os.unlink('test_write_associated.cif')
        # struct and audit_conform categories should be in associated file,
        # not the main file
        self.assertIn('_struct.title', assoc_file)
        self.assertNotIn('_struct.title', main_file)
        self.assertIn('_audit_conform.dict_name', assoc_file)
        self.assertNotIn('_audit_conform.dict_name', main_file)

    def test_write_associated_in_zip(self):
        """Test write() function with associated files in a ZipFile"""
        s = modelcif.System(id='system1')

        f = modelcif.associated.CIFFile(
            path='test_write_associated_in_zip.cif',
            categories=['struct', '_AUDIT_CONFORM'],
            entry_details='test details', entry_id='testcif')
        zf = modelcif.associated.ZipFile(path='t.zip', files=[f])
        r = modelcif.associated.Repository(url_root='https://example.com',
                                           files=[zf])
        s.repositories.append(r)

        fh = StringIO()
        modelcif.dumper.write(fh, [s])
        main_file = fh.getvalue()
        with open('test_write_associated_in_zip.cif') as fh:
            assoc_file = fh.read()
        os.unlink('test_write_associated_in_zip.cif')
        # struct and audit_conform categories should be in associated file,
        # not the main file
        self.assertIn('_struct.title', assoc_file)
        self.assertNotIn('_struct.title', main_file)
        self.assertIn('_audit_conform.dict_name', assoc_file)
        self.assertNotIn('_audit_conform.dict_name', main_file)

    def test_write_associated_copy(self):
        """Test write() function with associated files, copy_categories"""
        s = modelcif.System(id='system1')

        e1 = modelcif.Entity('ACGT')
        e1._id = 42
        s.entities.append(e1)

        f = modelcif.associated.CIFFile(
            path='/not/exist/foo.cif',
            local_path='test_write_associated_copy.cif',
            categories=['struct'], copy_categories=['entity', 'audit_conform'],
            entry_details='test details', entry_id='testcif')
        r = modelcif.associated.Repository(url_root='https://example.com',
                                           files=[f])
        s.repositories.append(r)

        fh = StringIO()
        modelcif.dumper.write(fh, [s])
        main_file = fh.getvalue()
        with open('test_write_associated_copy.cif') as fh:
            assoc_file = fh.read()
        os.unlink('test_write_associated_copy.cif')
        # struct category should be in associated file, not the main file
        self.assertIn('_struct.title', assoc_file)
        self.assertNotIn('_struct.title', main_file)
        # entity and audit conform categories should be in *both* files
        self.assertIn('_entity.type', assoc_file)
        self.assertIn('_entity.type', main_file)
        self.assertIn('_audit_conform.dict_name', assoc_file)
        self.assertIn('_audit_conform.dict_name', main_file)

    def test_write_associated_none(self):
        """Test write() function with associated files, no categories"""
        s = modelcif.System(id='system1')

        f = modelcif.associated.CIFFile(
            path='test_write_associated_none.cif')
        r = modelcif.associated.Repository(url_root='https://example.com',
                                           files=[f])
        s.repositories.append(r)

        fh = StringIO()
        modelcif.dumper.write(fh, [s])
        main_file = fh.getvalue()
        self.assertIn('_struct.title', main_file)
        self.assertIn('_audit_conform.dict_name', main_file)

    @unittest.skipIf(msgpack is None, "needs Python 3 and msgpack")
    def test_write_associated_binary(self):
        """Test write() function with associated binary files"""
        s = modelcif.System(id='system1')

        f = modelcif.associated.CIFFile(
            path='test_write_associated_binary.bcif',
            categories=['struct', '_AUDIT_CONFORM'],
            entry_details='test details', entry_id='testcif', binary=True)
        r = modelcif.associated.Repository(url_root='https://example.com',
                                           files=[f])
        s.repositories.append(r)

        fh = StringIO()
        modelcif.dumper.write(fh, [s])
        main_file = fh.getvalue()
        with open('test_write_associated_binary.bcif', 'rb') as fh:
            assoc_file = msgpack.unpack(fh, raw=False)
        os.unlink('test_write_associated_binary.bcif')
        assoc_cats = frozenset(
            x['name'] for x in assoc_file['dataBlocks'][0]['categories'])

        self.assertIn('_struct', assoc_cats)
        self.assertNotIn('_struct.title', main_file)
        self.assertIn('_audit_conform', assoc_cats)
        self.assertNotIn('_audit_conform.dict_name', main_file)

    def test_system_writer(self):
        """Test _SystemWriter utility class"""
        class BaseWriter:
            def flush(self):
                return 'flush called'

            def write_comment(self, comment):
                return 'write comment ' + comment

        s = modelcif.dumper._SystemWriter(BaseWriter(), {}, {})
        # These methods are not usually called in ordinary operation, but
        # we should provide them for Writer compatibility
        self.assertEqual(s.flush(), 'flush called')
        self.assertEqual(s.write_comment('foo'), 'write comment foo')

    def test_entity_non_poly_dumper(self):
        """Test EntityNonPolyDumper"""
        system = modelcif.System()
        # Polymeric entity (ignored)
        e1 = modelcif.Entity('ACGT')
        e1._id = 1
        e2 = ihm.Entity([ihm.NonPolymerChemComp('HEM')], description='heme')
        e2._id = 2
        e3 = ihm.Entity([ihm.NonPolymerChemComp('ZN')], description='zinc')
        e3._id = 3
        # Branched entity (ignored)
        e4 = modelcif.Entity(
            [ihm.SaccharideChemComp('NAG'), ihm.SaccharideChemComp('FUC')])
        system.entities.extend((e1, e2, e3, e4))

        t2 = modelcif.Template(e2, 'A', model_num=1, transformation=None)
        a1 = modelcif.AsymUnit(e1, 'foo')
        a2 = modelcif.NonPolymerFromTemplate(template=t2, explicit=True)
        system.asym_units.extend((a1, a2))

        dumper = modelcif.dumper._EntityNonPolyDumper()
        dumper.finalize(system)
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_pdbx_entity_nonpoly.entity_id
_pdbx_entity_nonpoly.name
_pdbx_entity_nonpoly.comp_id
_pdbx_entity_nonpoly.ma_model_mode
2 heme HEM explicit
3 zinc ZN .
#
""")

    def test_chem_comp_dumper(self):
        """Test ChemCompDumper"""
        system = modelcif.System()

        # ChemComp without ccd
        c1 = ihm.NonPolymerChemComp('C1', name='C1')

        # ChemComp using core CCD
        c2 = ihm.NonPolymerChemComp('C2', name='C2')
        c2.ccd = 'core'

        # ChemComp using MA CCD
        c3 = ihm.NonPolymerChemComp('C3', name='C3')
        c3.ccd = 'ma'

        # ChemComp with descriptors (local)
        c4 = ihm.NonPolymerChemComp('C4', name='C4')
        c4.ccd = None
        c4.descriptors = [modelcif.descriptor.IUPACName("foo")]

        e1 = modelcif.Entity(['A', 'C', c1, c2, c3, c4])
        system.entities.append(e1)

        e2 = modelcif.Entity('GT')
        t2 = modelcif.Template(e2, 'A', model_num=1, transformation=None)
        system.templates.append(t2)

        dumper = modelcif.dumper._ChemCompDumper()
        out = _get_dumper_output(dumper, system)
        # chem_comp should include both system.entities and system.templates
        self.assertEqual(out, """#
loop_
_chem_comp.id
_chem_comp.type
_chem_comp.name
_chem_comp.formula
_chem_comp.formula_weight
_chem_comp.ma_provenance
ALA 'L-peptide linking' ALANINE 'C3 H7 N O2' 89.093 'CCD Core'
C1 non-polymer C1 . . 'CCD Core'
C2 non-polymer C2 . . 'CCD Core'
C3 non-polymer C3 . . 'CCD MA'
C4 non-polymer C4 . . 'CCD local'
CYS 'L-peptide linking' CYSTEINE 'C3 H7 N O2 S' 121.158 'CCD Core'
GLY 'peptide linking' GLYCINE 'C2 H5 N O2' 75.067 'CCD Core'
THR 'L-peptide linking' THREONINE 'C4 H9 N O3' 119.119 'CCD Core'
#
""")

    def test_chem_comp_dumper_bad_ccd(self):
        """Test ChemCompDumper with invalid value for ccd"""
        system = modelcif.System()

        c1 = ihm.NonPolymerChemComp('C1', name='C1')
        c1.ccd = 'garbage'

        e1 = modelcif.Entity([c1])
        system.entities.append(e1)

        dumper = modelcif.dumper._ChemCompDumper()
        self.assertRaises(KeyError, _get_dumper_output, dumper, system)
        # Should work with checks disabled
        _ = _get_dumper_output(dumper, system, check=False)

    def test_chem_comp_descriptor_dumper(self):
        """Test ChemCompDescriptorDumper"""
        class MockObject:
            pass

        system = modelcif.System()

        # Old-style ChemComp without descriptors
        c1 = ihm.NonPolymerChemComp('C1', name='C1name')
        if hasattr(c1, 'descriptor'):
            del c1.descriptors

        c2 = ihm.NonPolymerChemComp('C2', name='C2name')
        c2.ccd = None
        soft = MockObject()
        soft._id = 42
        c2.descriptors = [modelcif.descriptor.IUPACName("foo"),
                          modelcif.descriptor.PubChemCID(
                              "bar", details="test details", software=soft)]

        e1 = modelcif.Entity(['A', 'C', c1, c2])
        system.entities.append(e1)

        dumper = modelcif.dumper._ChemCompDescriptorDumper()
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_ma_chem_comp_descriptor.ordinal_id
_ma_chem_comp_descriptor.chem_comp_id
_ma_chem_comp_descriptor.chem_comp_name
_ma_chem_comp_descriptor.type
_ma_chem_comp_descriptor.value
_ma_chem_comp_descriptor.details
_ma_chem_comp_descriptor.software_id
1 C2 C2name 'IUPAC Name' foo . .
2 C2 C2name 'PubChem CID' bar 'test details' 42
#
""")

    def test_struct_ref(self):
        """Test StructRefDumper"""
        system = ihm.System()
        lpep = ihm.LPeptideAlphabet()
        sd = modelcif.reference.SeqDif(
            seq_id=2, db_monomer=lpep['W'],
            monomer=lpep['S'], details='Test mutation')
        r1 = modelcif.reference.UniProt(
            code='NUP84_YEAST', accession='P52891', sequence='MELWPTYQT',
            details='test sequence')
        r1.alignments.append(modelcif.reference.Alignment(
            db_begin=3, seq_dif=[sd]))
        r2 = modelcif.reference.UniProt(
            code='testcode', accession='testacc', sequence='MELSPTYQT',
            details='test2')
        r2.alignments.append(modelcif.reference.Alignment(
            db_begin=4, db_end=5, entity_begin=2, entity_end=3))
        r2.alignments.append(modelcif.reference.Alignment(
            db_begin=9, db_end=9, entity_begin=4, entity_end=4))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            r3 = modelcif.reference.UniProt(
                code='testcode2', accession='testacc2', sequence=None)
        r3.alignments.append(modelcif.reference.Alignment(
            db_begin=4, db_end=5, entity_begin=2, entity_end=3))
        r4 = modelcif.reference.UniProt(
            code='testcode3', accession='testacc3', sequence=ihm.unknown)
        r4.alignments.append(modelcif.reference.Alignment(
            db_begin=4, db_end=5, entity_begin=2, entity_end=3))
        system.entities.append(modelcif.Entity(
            'LSPT', references=[r1, r2, r3, r4]))
        dumper = ihm.dumper._EntityDumper()
        dumper.finalize(system)  # Assign entity IDs

        dumper = ihm.dumper._StructRefDumper()
        dumper.finalize(system)  # Assign IDs
        out = _get_dumper_output(dumper, system)
        self.assertEqual(out, """#
loop_
_struct_ref.id
_struct_ref.entity_id
_struct_ref.db_name
_struct_ref.db_code
_struct_ref.pdbx_db_accession
_struct_ref.pdbx_align_begin
_struct_ref.pdbx_seq_one_letter_code
_struct_ref.details
1 1 UNP NUP84_YEAST P52891 3 LWPTYQT 'test sequence'
2 1 UNP testcode testacc 4 SPTYQT test2
3 1 UNP testcode2 testacc2 4 . .
4 1 UNP testcode3 testacc3 4 ? .
#
#
loop_
_struct_ref_seq.align_id
_struct_ref_seq.ref_id
_struct_ref_seq.seq_align_beg
_struct_ref_seq.seq_align_end
_struct_ref_seq.db_align_beg
_struct_ref_seq.db_align_end
1 1 1 4 3 6
2 2 2 3 4 5
3 2 4 4 9 9
4 3 2 3 4 5
5 4 2 3 4 5
#
#
loop_
_struct_ref_seq_dif.pdbx_ordinal
_struct_ref_seq_dif.align_id
_struct_ref_seq_dif.db_mon_id
_struct_ref_seq_dif.pdbx_seq_db_seq_num
_struct_ref_seq_dif.mon_id
_struct_ref_seq_dif.seq_num
_struct_ref_seq_dif.details
1 1 TRP ? SER 2 'Test mutation'
#
""")

    def test_write_branched(self):
        """Test write() output of branched entity information"""
        # Tests for individual dumpers are already present in python-ihm;
        # here we just test to make sure they are being called by write()
        s = modelcif.System()
        e = modelcif.Entity(
            [ihm.SaccharideChemComp('NAG'),
             ihm.SaccharideChemComp('BMC'),
             ihm.SaccharideChemComp('FUC')])
        e.branch_descriptors.append(
            ihm.BranchDescriptor('bar', type='typ2'))
        e.branch_links.append(
            ihm.BranchLink(num1=2, atom_id1='CA', leaving_atom_id1='H1',
                           num2=3, atom_id2='N', leaving_atom_id2='H2'))
        s.entities.append(e)
        asym = modelcif.AsymUnit(e, 'foo')
        s.asym_units.append(asym)

        fh = StringIO()
        modelcif.dumper.write(fh, [s])
        output = fh.getvalue()
        self.assertIn('_pdbx_entity_branch_list.hetero', output)
        self.assertIn('_pdbx_entity_branch.type', output)
        self.assertIn('_pdbx_branch_scheme.entity_id', output)
        self.assertIn('_pdbx_entity_branch_descriptor.ordinal', output)
        self.assertIn('_pdbx_entity_branch_link.comp_id_1', output)


if __name__ == '__main__':
    unittest.main()



import ihm.representation
from ihm.model import Atom, ModelGroup  # noqa: F401
import modelcif.data
from ihm.util import _check_residue_range


# Provide ma-specific docs for Atom
Atom.__doc__ = """Coordinates of part of the model represented by an atom.

See :meth:`Model.get_atoms` for more details.

:param asym_unit: The asymmetric unit that this atom represents
:type asym_unit: :class:`modelcif.AsymUnit`
:param int seq_id: The residue index represented by this atom
       (can be None for HETATM sites)
:param str atom_id: The name of the atom in the residue
:param str type_symbol: Element name
:param float x: x coordinate of the atom
:param float y: y coordinate of the atom
:param float z: z coordinate of the atom
:param bool het: True for HETATM sites, False (default) for ATOM
:param float biso: Temperature factor or equivalent (if applicable)
:param float occupancy: Fraction of the atom type present
       (if applicable)
"""

# Provide ma-specific docs for ModelGroup
ModelGroup.__doc__ = """A set of related models. See :class:`Model`.
It is implemented as a simple list of the models.

These objects are typically stored directly in the system; see
:attr:`modelcif.System.model_groups`.

:param elements: Initial set of models in the group.
:param str name: Descriptive name for the group.
"""


class Model(modelcif.data.Data):
    """Base class for coordinates of a single structure.
       Use a subclass such as :class:`HomologyModel` or
       :class:`AbInitioModel`, or represent a custom model type by
       creating a new subclass and providing a docstring to describe it, e.g.::

           class CustomModel(Model):
               "custom model type"

       :param assembly: The :class:`modelcif.AsymUnit` objects that make up
              this model.
       :type assembly: :class:`modelcif.Assembly`
       :param str name: Short name for this model.
    """
    data_content_type = 'model coordinates'
    model_type = "Other"

    def __init__(self, assembly, name=None):
        super().__init__(name)
        self.assembly = assembly
        # Assume everything is atomic for ModelCIF models
        self.representation = ihm.representation.Representation(
            [ihm.representation.AtomicSegment(seg, rigid=False)
             for seg in assembly])
        self._atoms = []

        #: List of residue ranges that were explicitly not modeled. See
        #: :class:`NotModeledResidueRange`.
        self.not_modeled_residue_ranges = []

        #: Quality scores for the model or part of it (a simple list of
        #: metric objects; see :mod:`modelcif.qa_metric`)
        self.qa_metrics = []

    def _get_other_details(self):
        if (type(self) is not Model
                and self.model_type == Model.model_type):
            return self.__doc__.split('\n')[0]

    other_details = property(
        _get_other_details,
        doc="More information about a custom model type. "
            "By default it is the first line of the docstring.")

    def get_atoms(self):
        """Yield :class:`Atom` objects that represent this model.

           The default implementation simply iterates over an internal
           list of atoms, but this is not very memory-efficient, particularly
           if the atoms are already stored somewhere else, e.g. in the
           software's own data structures. It is recommended to subclass
           and provide a more efficient implementation. For example,
           `the modbase_pdb_to_cif script <https://github.com/salilab/modbase_utils/blob/main/modbase_pdb_to_cif.py>`_
           uses a custom ``MyModel`` subclass that creates Atom objects on
           the fly from PDB ATOM or HETATM lines.
        """  # noqa: E501
        for a in self._atoms:
            yield a

    def add_atom(self, atom):
        self._atoms.append(atom)


class HomologyModel(Model):
    """Coordinates of a single structure generated using homology
       or comparative modeling.

       See :class:`Model` for a description of the parameters.
    """
    model_type = "Homology model"
    other_details = None


class AbInitioModel(Model):
    """Coordinates of a single structure generated using ab initio modeling.

       See :class:`Model` for a description of the parameters.
    """
    model_type = "Ab initio model"
    other_details = None


class NotModeledResidueRange:
    """A range of residues that were explicitly not modeled.
       See :attr:`Model.not_modeled_residue_ranges`.
       These ranges are not explicitly stored in the mmCIF file,
       but they will be excluded from the ``pdbx_poly_seq_scheme`` table.

       :param asym_unit: The asymmetric unit to which the residues belong.
       :type asym_unit: :class:`~modelcif.AsymUnit`
       :param int seq_id_begin: Starting residue in the range.
       :param int seq_id_end: Ending residue in the range.
    """
    def __init__(self, asym_unit, seq_id_begin, seq_id_end):
        self.asym_unit = asym_unit
        self.seq_id_begin, self.seq_id_end = seq_id_begin, seq_id_end
        _check_residue_range((seq_id_begin, seq_id_end), asym_unit.entity)



import utils
import os
import unittest
import sys
import shutil
import subprocess
try:
    import msgpack
except ImportError:
    msgpack = None

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)

import modelcif.reader


def get_example_dir():
    return os.path.join(TOPDIR, "examples")


def get_example_path(fname):
    return os.path.join(get_example_dir(), fname)


class Tests(unittest.TestCase):

    @unittest.skipIf('APPVEYOR' in os.environ,
                     "AppVeyor environments have old SSL certs")
    @unittest.skipIf('GITHUB_ACTIONS' in os.environ,
                     "Example is slow and fails when ModBase is down")
    def test_validate_modbase_example(self):
        """Test validate_modbase example"""
        subprocess.check_call([sys.executable,
                               get_example_path("validate_modbase.py")])

    @unittest.skipIf('APPVEYOR' in os.environ,
                     "AppVeyor environments have old SSL certs")
    @unittest.skipIf('GITHUB_ACTIONS' in os.environ,
                     "Example is slow and fails when ModBase is down")
    def test_validate_mmcif_example(self):
        """Test validate_mmcif example"""
        with utils.temporary_directory() as tmpdir:
            subprocess.check_call([sys.executable,
                                   get_example_path("validate_mmcif.py"),
                                   get_example_path("input/ligands.cif")],
                                  cwd=tmpdir)

    def test_associated_example(self):
        """Test associated example"""
        subprocess.check_call([sys.executable,
                               get_example_path("associated.py")])

    def test_mkmodbase_example(self):
        """Test mkmodbase example"""
        with utils.temporary_directory() as tmpdir:
            subprocess.check_call([sys.executable,
                                   get_example_path("mkmodbase.py")],
                                  cwd=tmpdir)

            # Make sure that a complete output file was produced and that we
            # can read it
            with open(os.path.join(tmpdir, 'output.cif')) as fh:
                contents = fh.readlines()
            self.assertEqual(len(contents), 454)
            with open(os.path.join(tmpdir, 'output.cif')) as fh:
                s, = modelcif.reader.read(fh)

    def test_ligands_example(self):
        """Test ligands example"""
        with utils.temporary_directory() as tmpdir:
            subprocess.check_call([sys.executable,
                                   get_example_path("ligands.py")],
                                  cwd=tmpdir)

            # Make sure that a complete output file was produced and that we
            # can read it
            with open(os.path.join(tmpdir, 'output.cif')) as fh:
                contents = fh.readlines()
            self.assertEqual(len(contents), 334)
            with open(os.path.join(tmpdir, 'output.cif')) as fh:
                s, = modelcif.reader.read(fh)

    @unittest.skipIf(msgpack is None, "BinaryCIF needs msgpack")
    def test_convert_bcif_example(self):
        """Test convert_bcif example"""
        with utils.temporary_directory() as tmpdir:
            from_input = get_example_path("input")
            to_input = os.path.join(tmpdir, 'input')
            os.mkdir(to_input)
            shutil.copy(os.path.join(from_input, "ligands.cif"), to_input)
            subprocess.check_call([sys.executable,
                                   get_example_path("convert_bcif.py")],
                                  cwd=tmpdir)

            # Make sure that a complete output file was produced and that we
            # can read it
            with open(os.path.join(tmpdir, 'ligands.bcif'), 'rb') as fh:
                s, = modelcif.reader.read(fh, format='BCIF')


if __name__ == '__main__':
    unittest.main()



# This example demonstrates reading ModelCIF "associated" files.
# Some repositories, such as ModelArchive, split the deposited data into
# multiple mmCIF files, putting some of the quality score information not
# into the main mmCIF file but into a separate "associated" file.
# These associated files are referenced in the main file
# (see System.repositories) so we can programmatically download and
# extract them.
# This example requires Python 3.

import urllib.request
import zipfile
import tempfile
import shutil
import modelcif.reader


# Get any associated files containing pairwise QA scores
def _get_zip_scores_files(s):
    for repo in s.repositories:
        for f in repo.files:
            if isinstance(f, modelcif.associated.ZipFile):
                for zf in f.files:
                    if isinstance(
                            zf, modelcif.associated.QAMetricsFile):
                        yield zf, f, repo


# Download entry ma-bak-cepc-0944 directly from ModelArchive
url = "https://www.modelarchive.org/doi/10.5452/ma-bak-cepc-0944.cif"
with urllib.request.urlopen(url) as fh:
    s, = modelcif.reader.read(fh)


# Get any referenced associated files containing QA scores. For ModelArchive,
# these are stored in an mmCIF file that is then compressed into a zip file
for scores, archive, repo in _get_zip_scores_files(s):
    url = repo.get_url(archive)
    # Download the referenced zip file directly from ModelArchive
    with urllib.request.urlopen(repo.get_url(archive)) as f_url:
        with tempfile.NamedTemporaryFile() as f_zip:
            shutil.copyfileobj(f_url, f_zip)
            # Extract the scores file from the zip file
            with zipfile.ZipFile(f_zip) as zf:
                with zf.open(scores.path) as f_scores:
                    # Add scores in the file to our existing System
                    modelcif.reader.read(f_scores, add_to_system=s)

for mg in s.model_groups:
    for m in mg:
        print("Model %s contains %d QA metrics" % (m, len(m.qa_metrics)))



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.associated


class Tests(unittest.TestCase):
    def test_local_pairwise_qa_scores_file(self):
        """Test LocalPairwiseQAScoresFile class"""
        self.assertWarns(UserWarning,
                         modelcif.associated.LocalPairwiseQAScoresFile,
                         path='foo')


if __name__ == '__main__':
    unittest.main()



# This example demonstrates the use of the Python IHM library's validator
# to validate a user-provided input mmCIF file.
# See also validate_modbase.py for a more detailed example.

import sys
import ihm.dictionary
import urllib.request


if len(sys.argv) != 2:
    print("Usage: %s input.cif" % sys.argv[0], file=sys.stderr)
    sys.exit(1)
fname = sys.argv[1]

# Read in the ModelCIF and PDBx dictionary from https://mmcif.wwpdb.org/
with urllib.request.urlopen(
        'https://mmcif.wwpdb.org/dictionaries/ascii/mmcif_ma.dic') as fh:
    pdbx_mcif = ihm.dictionary.read(fh)

# Validate the mmCIF file assuming it is UTF8 encoded.
# See validate_modbase.py for code to fallback to ASCII for non-UTF8 files.
with open(fname, encoding='UTF-8') as fh:
    pdbx_mcif.validate(fh)

# Similarly, to validate a BinaryCIF file, use:
# with open(fname, 'rb') as fh:
#    pdbx_mcif.validate(fh, format='BCIF')



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.model


class Tests(unittest.TestCase):
    def test_model(self):
        """Test Model classes"""
        m = modelcif.model.HomologyModel([])
        self.assertEqual(m.model_type, "Homology model")
        self.assertIsNone(m.other_details)

        # generic "other" model
        m = modelcif.model.Model([])
        self.assertEqual(m.model_type, "Other")
        self.assertIsNone(m.other_details)

        # custom "other" model
        class CustomRef(modelcif.model.Model):
            """foo
               bar"""

        m = CustomRef([])
        self.assertEqual(m.model_type, "Other")
        self.assertEqual(m.other_details, "foo")


if __name__ == '__main__':
    unittest.main()



import unittest
import modelcif.reader
import modelcif.dumper
import urllib.request
import os


class Tests(unittest.TestCase):
    def _read_cif(self, url):
        with urllib.request.urlopen(url) as fh:
            s, = modelcif.reader.read(fh)
        return s

    def _write_cif(self, s, check=True):
        with open('test.cif', 'w') as fh:
            modelcif.dumper.write(fh, [s], check=check)
        os.unlink('test.cif')

    def test_modbase(self):
        """Test ModBase structure without errors"""
        model_id = '3c79945a94ec00cac8a03104e853ca50'
        modbase_top = 'https://salilab.org/modbase/retrieve/modbase'
        url = '%s/?modelID=%s&format=mmcif' % (modbase_top, model_id)
        s = self._read_cif(url)
        self._write_cif(s)

    def test_swiss_model(self):
        """Test SWISS-MODEL structure without errors"""
        model_id = '680335e5cca47f7d2b00afc1'
        url = 'https://swissmodel.expasy.org/repository/%s.cif' % model_id
        s = self._read_cif(url)
        self._write_cif(s)

    def test_alpha_fold(self):
        """Test AlphaFold structure without errors"""
        model_id = 'AF-B4GKE9-F1-model_v6'
        url = 'https://alphafold.ebi.ac.uk/files/%s.cif' % model_id
        s = self._read_cif(url)
        self._write_cif(s)


if __name__ == '__main__':
    unittest.main()



# This example demonstrates using the library to convert an mmCIF file
# containing ModelCIF data to BinaryCIF format.

# Import used classes.
import modelcif
import modelcif.dumper
import modelcif.reader

# Read in an existing mmCIF file:
with open('input/ligands.cif') as fh:
    systems = modelcif.reader.read(fh, format='mmCIF')

# Write a new BinaryCIF file containing the same data:
with open('ligands.bcif', 'wb') as fh:
    modelcif.dumper.write(fh, systems, format='BCIF')



#!/usr/bin/env python3

"""
Add minimal ModelCIF-related tables to an mmCIF file.

Given any mmCIF file as input, this script will add any missing
ModelCIF-related tables and write out a new file that is minimally compliant
with the ModelCIF dictionary.

This is done by simply reading in the original file with python-modelcif and
then writing it out again, so
  a) any data in the input file that is not understood by python-modelcif
     will be lost on output; and
  b) input files that aren't compliant with the PDBx dictionary, or that
     contain syntax errors or other problems, may crash or otherwise confuse
     python-modelcif.

While a best effort is made, it is not guaranteed that the output file is
valid. It is recommended that it is run through a validator such as
examples/validate_mmcif.py and any errors corrected or reported as
issues.
"""


import modelcif.reader
import modelcif.dumper
import modelcif.model
import ihm.util
import os
import argparse


def add_modelcif_info(s):
    if not s.title:
        s.title = 'Auto-generated system'
    if not s.protocols:
        default_protocol = modelcif.protocol.Protocol()
        step = modelcif.protocol.ModelingStep(
            name='modeling', input_data=None, output_data=None)
        default_protocol.steps.append(step)
        s.protocols.append(default_protocol)

    for model_group in s.model_groups:
        for model in model_group:
            # Entity description is also used by python-modelcif for
            # ma_data.name, which is mandatory, so it cannot be unknown/?
            for asym in model.assembly:
                if asym.entity.description is ihm.unknown:
                    asym.entity.description = "target"

            model.not_modeled_residue_ranges.extend(
                _get_not_modeled_residues(model))
    return s


def _get_not_modeled_residues(model):
    """Yield NotModeledResidueRange objects for all residue ranges in the
       Model that are not referenced by Atom objects"""
    for assem in model.assembly:
        asym = assem.asym if hasattr(assem, 'asym') else assem
        if not asym.entity.is_polymeric():
            continue
        # Make a set of all residue indices of this asym "handled"
        # by being modeled with Atom objects
        handled_residues = set()
        for atom in model._atoms:
            if atom.asym_unit is asym:
                handled_residues.add(atom.seq_id)
        # Convert set to a list of residue ranges
        handled_residues = ihm.util._make_range_from_list(
            sorted(handled_residues))
        # Return not-modeled for each non-handled range
        for r in ihm.util._invert_ranges(handled_residues,
                                         end=assem.seq_id_range[1],
                                         start=assem.seq_id_range[0]):
            yield modelcif.model.NotModeledResidueRange(asym, r[0], r[1])


def get_args():
    p = argparse.ArgumentParser(
        description="Add minimal ModelCIF-related tables to an mmCIF file.")
    p.add_argument("input", metavar="input.cif", help="input mmCIF file name")
    p.add_argument("output", metavar="output.cif",
                   help="output mmCIF file name",
                   default="output.cif", nargs="?")
    return p.parse_args()


def main():
    args = get_args()

    if (os.path.exists(args.input) and os.path.exists(args.output)
            and os.path.samefile(args.input, args.output)):
        raise ValueError("Input and output are the same file")

    with open(args.input) as fh:
        with open(args.output, 'w') as fhout:
            modelcif.dumper.write(
                fhout,
                [add_modelcif_info(s) for s in modelcif.reader.read(fh)])


if __name__ == '__main__':
    main()



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import sys

VERSION = "1.8"

copy_args = sys.argv[1:]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='modelcif',
      version=VERSION,
      script_args=copy_args,
      description='Package for handling ModelCIF mmCIF and BinaryCIF files',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Ben Webb',
      author_email='benmwebb@gmail.com',
      url='https://github.com/ihmwg/python-modelcif',
      packages=['modelcif', 'modelcif.util'],
      install_requires=['ihm>=2.11'],
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Intended Audience :: Science/Research",
          "Topic :: Scientific/Engineering",
      ])
